from enum import Enum
import json
from cryptography.hazmat.primitives.asymmetric import ed25519
from typing import Literal, TypedDict, Dict, List, Union

from ..types import JSONType
from ..api import Api, Endpoints

from decentriq_dcr_compiler.schemas.data_room import (
    DataRoomComputeAction as DataRoomComputeActionSchema,
    AuditLogLine,
)


class SignedContent(TypedDict):
    id: str
    content: str
    signatureHex: str
    version: int
    casIndex: int


class DataRoom(TypedDict):
    id: str
    ownerId: str
    isStopped: bool
    signedContent: SignedContent


class Action(TypedDict):
    id: str
    data_room_id: str
    jobs: List[str]


class Job(TypedDict):
    id: str


class TaskView(TypedDict):
    name: str
    results: Dict[str, str]
    dependencies: List[str]
    completed: List[str]


class JobTasksView(TypedDict):
    tasks: Dict[str, TaskView]
    target_task: str


class JobStatus(str, Enum):
    RUNNING = "Running"
    SUCCESS = "Success"
    FAILED = "Failed"


class JobStatusView(TypedDict):
    status: JobStatus
    registered_at: str
    completed_at: str


class RawError(TypedDict):
    task_name: str
    error_message: str


class RawErrorsView(TypedDict):
    raw_errors: List[RawError]


class DataRoomAuditLogUnverified(TypedDict):
    signed_content: SignedContent


class DataRoomAuditLog(TypedDict):
    logs: List[AuditLogLine]


class DataRoomActionSummary(TypedDict):
    action: str
    jobs: List[str]


class DataRoomActionsView(TypedDict):
    actions: Dict[str, DataRoomActionSummary]


class DataLabCompatibilityResponse(TypedDict):
    errorsByDataLab: Dict[str, List[str]]


class PolicyStatus(str, Enum):
    OK = "ok"
    NOT_FOUND = "not_found"
    NOT_PROVISIONED = "not_provisioned"


class PolicyNotFound(TypedDict):
    status: Literal[PolicyStatus.NOT_FOUND]


class PolicyNotProvisioned(TypedDict):
    status: Literal[PolicyStatus.NOT_PROVISIONED]


class PolicyMetadata(TypedDict):
    status: Literal[PolicyStatus.OK]
    manifest_hash_hex: str
    target_leaf: str


Policy = Union[PolicyNotFound, PolicyNotProvisioned, PolicyMetadata]


class ClientV2:
    def __init__(self, api: Api):
        self.api = api

    """
    Data Room API
    """

    def create_data_room(
        self, create_data_room: JSONType, verification_key: bytes
    ) -> DataRoom:
        body = {"payload": json.dumps(create_data_room)}
        response = self.api.post(
            endpoint=Endpoints.DATA_ROOMS_COLLECTION,
            req_body=json.dumps(body),
            headers={"Content-type": "application/json"},
        )
        if response.status_code == 201 and "Location" in response.headers:
            location = response.headers["Location"]
            data_room_id = location.split("/")[-1]
            return self.get_data_room(data_room_id, verification_key)
        else:
            raise Exception(
                f"Failed to create data room: {response.status_code} {response.text}"
            )

    def get_data_room(self, data_room_id: str, verification_key: bytes) -> DataRoom:
        endpoint = Endpoints.DATA_ROOM.replace(":data_room_id", data_room_id)
        unverified_data_room: DataRoom = self.api.get(
            endpoint=endpoint,
        ).json()
        return self._verify_data_room(unverified_data_room, verification_key)

    def _verify_data_room(
        self, unverified_data_room: DataRoom, verification_key: bytes
    ) -> DataRoom:
        verified_signed_content = self._get_verified_signed_content(
            unverified_data_room["id"],
            "dcr_state",
            unverified_data_room["signedContent"],
            verification_key,
        )
        verified_data_room = unverified_data_room
        verified_data_room["signedContent"]["content"] = verified_signed_content[
            "content"
        ]
        return verified_data_room

    def get_data_room_audit_log(
        self, data_room_id: str, verification_key: bytes
    ) -> DataRoomAuditLog:
        endpoint = Endpoints.DATA_ROOM_AUDIT_LOG.replace(":data_room_id", data_room_id)
        response: DataRoomAuditLogUnverified = self.api.get(
            endpoint=endpoint,
        ).json()
        verified_audit_log = self._verify_audit_log(
            data_room_id, response, verification_key
        )
        return verified_audit_log

    def _verify_audit_log(
        self,
        dcr_id: str,
        unverified_audit_log: DataRoomAuditLogUnverified,
        verification_key: bytes,
    ) -> DataRoomAuditLog:
        signed_content = self._get_verified_signed_content(
            dcr_id,
            "dcr_audit_log",
            unverified_audit_log["signed_content"],
            verification_key,
        )
        content = json.loads(signed_content["content"])
        return DataRoomAuditLog(logs=[AuditLogLine.model_validate(c) for c in content])

    def _get_verified_signed_content(
        self,
        dcr_id: str,
        content_type: str,
        signed_content: SignedContent,
        verification_key: bytes,
    ) -> SignedContent:
        dcr_id_bytes = dcr_id.encode("utf-8")
        version_bytes = signed_content["version"].to_bytes(4, "little")
        content_type_bytes = content_type.encode("utf-8")
        content_bytes = signed_content["content"].encode("utf-8")
        # Concatenate bytes in same order as Rust
        sign_over_payload = (
            dcr_id_bytes + version_bytes + content_type_bytes + content_bytes
        )
        # Verify using ed25519
        public_key = ed25519.Ed25519PublicKey.from_public_bytes(verification_key)
        try:
            public_key.verify(
                bytes.fromhex(signed_content["signatureHex"]), sign_over_payload
            )
            return signed_content
        except Exception:
            raise Exception("Invalid signed content")

    def get_data_room_actions(self, data_room_id: str) -> DataRoomActionsView:
        endpoint = Endpoints.DATA_ROOM_ACTIONS.replace(":data_room_id", data_room_id)
        response: DataRoomActionsView = self.api.get(
            endpoint=endpoint,
        ).json()
        return response

    def get_data_room_policies(
        self, data_room_id: str, policy_ids: List[str]
    ) -> Dict[str, Policy]:
        endpoint = Endpoints.DATA_ROOM_POLICIES.replace(":data_room_id", data_room_id)
        response: Dict[str, Policy] = self.api.post(
            endpoint=endpoint,
            req_body=json.dumps(policy_ids),
            headers={"Content-type": "application/json"},
        ).json()
        return response

    def stop_data_room(self, data_room_id: str) -> None:
        endpoint = Endpoints.DATA_ROOM_STOP.replace(":data_room_id", data_room_id)
        self.api.post(
            endpoint=endpoint,
        )

    """
    Action API
    """

    def get_data_room_action_info(
        self, data_room_id: str, data_room_compute_action: JSONType
    ) -> Action:
        body = {
            "payload": data_room_compute_action,
            "data_room_id": data_room_id,
        }
        response: Action = self.api.post(
            endpoint=Endpoints.ACTIONS_COLLECTION,
            req_body=json.dumps(body),
            headers={"Content-type": "application/json"},
        ).json()
        return response

    """
    Job API
    """

    def create_job(self, data_room_id: str, data_room_compute_action: JSONType) -> Job:
        body = {
            "payload": json.dumps(data_room_compute_action),
            "data_room_id": data_room_id,
        }
        response = self.api.post(
            endpoint=Endpoints.JOBS_COLLECTION,
            req_body=json.dumps(body),
            headers={"Content-type": "application/json"},
        )
        if response.status_code == 201 and "Location" in response.headers:
            location = response.headers["Location"]
            job_id = location.split("/")[-1]
            return self.get_job(job_id)
        else:
            raise Exception(
                f"Failed to create job: {response.status_code} {response.text}"
            )

    def get_job(self, job_id: str) -> Job:
        endpoint = Endpoints.JOB.replace(":job_id", job_id)
        response: Job = self.api.get(
            endpoint=endpoint,
        ).json()
        return response

    def get_job_tasks(self, job_id: str) -> JobTasksView:
        endpoint = Endpoints.JOB_TASKS_COLLECTION.replace(":job_id", job_id)
        response: JobTasksView = self.api.get(
            endpoint=endpoint,
        ).json()
        return response

    def get_job_status(self, job_id: str) -> JobStatusView:
        endpoint = Endpoints.JOB_STATUS.replace(":job_id", job_id)
        response: JobStatusView = self.api.get(
            endpoint=endpoint,
        ).json()
        return response

    def get_job_freshness(self, job_id: str) -> bool:
        endpoint = Endpoints.JOB_FRESHNESS.replace(":job_id", job_id)
        response: bool = self.api.get(
            endpoint=endpoint,
        ).json()
        return response

    def download_job_raw_errors(self, job_id: str) -> RawErrorsView:
        endpoint = Endpoints.JOB_RAW_ERRORS_RESULT.replace(":job_id", job_id)
        response: RawErrorsView = self.api.get(endpoint=endpoint).json()
        return response

    """
    Compatibility API
    """

    def check_data_labs_compatibility(
        self, data_room_id: str, data_lab_ids: List[str]
    ) -> DataLabCompatibilityResponse:
        body = {"mediaDcrId": data_room_id, "dataLabIds": data_lab_ids}
        response: DataLabCompatibilityResponse = self.api.post(
            endpoint=Endpoints.CHECK_DATA_LABS_COMPATIBILITY,
            req_body=json.dumps(body),
            headers={"Content-type": "application/json"},
        ).json()
        return response



import io, json
from typing import BinaryIO

class SnowflakeCredentials:
    def __init__(self, account_id: str, username: str, password: str) -> None:
        self.account_id = account_id
        self.username = username
        self.password = password

    def as_binary_io(self) -> BinaryIO:
        credentials = {
            "accountId": self.account_id,
            "username": self.username,
            "password": self.password,
        }
        return io.BytesIO(json.dumps(credentials).encode())


import json
from typing import Dict

from ...session import Session
from ...storage import Key
from ...types import EnclaveSpecification
from ..builders import DataRoomCommitBuilder
from ..container import read_result_as_zipfile
from ..permission import Permissions
from . import DatasetSink
from .proto import FileSelection, SingleFile, SinkInput, ZipFile


def store_computation_result_as_dataset(
    session: Session,
    enclave_specs: Dict[str, EnclaveSpecification],
    key: Key,
    dataset_name: str,
    data_room_id: str,
    compute_node_id: str,
):
    current_config, history_pin = session.retrieve_current_data_room_configuration(
        data_room_id
    )
    builder = DataRoomCommitBuilder(
        "dataset_sink_node",
        data_room_id,
        current_configuration=current_config,
        history_pin=history_pin,
        enclave_specs=enclave_specs,
    )
    sink_node_id = f"{compute_node_id}_sink"
    key_node_id = f"{sink_node_id}_key"

    inputs = [
        SinkInput(
            name=dataset_name,
            dependency=compute_node_id,
            zip=ZipFile(files=FileSelection(names=[SingleFile(name="hello.txt")])),
        )
    ]

    sink = DatasetSink(
        name=sink_node_id,
        inputs=inputs,
        encryption_key_dependency=f"{compute_node_id}_sink_key",
        is_key_hex_encoded=True,
    )
    builder.add_compute_node(sink, node_id=sink_node_id)
    builder.add_parameter_node(key_node_id, is_required=True, node_id=key_node_id)

    builder.add_user_permission(
        email=session.auth.user_id,
        authentication_method=session.client.decentriq_pki_authentication,
        permissions=[
            Permissions.execute_compute(sink_node_id),
            Permissions.retrieve_compute_result(sink_node_id),
        ],
    )
    commit = builder.build()
    commit_id = session.publish_data_room_configuration_commit(commit)

    job_id = session.run_dev_computation(
        data_room_id,
        commit_id,
        sink_node_id,
        parameters={key_node_id: key.material.hex()},
    )
    results = session.get_computation_result(job_id)
    results_zip = read_result_as_zipfile(results)
    datasets_meta_str = results_zip.read("datasets.json").decode()
    datasets_meta = json.loads(datasets_meta_str)

    return datasets_meta["datasets"][0]["manifestHash"]



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Dict, List, Optional

from decentriq_dcr_compiler.schemas import SqliteComputationNode
from typing_extensions import Self

from ..session import Session
from .high_level_node import StructuredOutputNode
from .node_definitions import NodeDefinition

if TYPE_CHECKING:
    from ..client import Client


class SqliteComputeNodeDefinition(NodeDefinition):
    """
    Class representing an SQLite Computation Node Definition.
    """

    def __init__(
        self,
        name: str,
        query: str,
        dependencies: Optional[List[str]] = None,
        enable_logs_on_error: bool = False,
        enable_logs_on_success: bool = False,
        id: Optional[str] = None,
    ) -> None:
        """
        Initialise a `SqliteComputeNodeDefinition`:

        **Parameters**:
        - `name`: Name of the `SqliteComputeNodeDefinition`.
        - `query`: SQLite query string.
        - `dependencies`: Mappings between node id and the table name under
          which they should be made available.
        - `enable_logs_on_error`: Enable logs in the event of an error.
        - `enable_logs_on_success`: Enable logs when the computation is successful.
        - `id`: Optional ID of the `SqliteComputeNodeDefinition`.
        """
        super().__init__(name=name, id=id or name)
        self.query = query
        self.dependencies = dependencies
        self.enable_logs_on_error = enable_logs_on_error
        self.enable_logs_on_success = enable_logs_on_success
        # SQLite computations use the Python worker under the hood.
        self.specification_id = "decentriq.python-ml-worker-32-64"
        self.static_content_specification_id = "decentriq.driver"

    @property
    def required_workers(self):
        return [self.static_content_specification_id, self.specification_id]

    def _get_high_level_representation(self) -> Dict[str, str]:
        """
        Retrieve the high level representation of the `SqliteComputeNodeDefinition`.
        """
        dependencies = []
        if self.dependencies:
            for dependency in self.dependencies:
                dependencies.append(
                    {
                        "nodeId": dependency,
                        "tableName": dependency,
                    }
                )
        computation_node = {
            "id": self.id,
            "name": self.name,
            "kind": {
                "computation": {
                    "kind": {
                        "sqlite": {
                            "dependencies": dependencies,
                            "enableLogsOnError": self.enable_logs_on_error,
                            "enableLogsOnSuccess": self.enable_logs_on_success,
                            "sqliteSpecificationId": self.specification_id,
                            "statement": self.query,
                            "staticContentSpecificationId": self.static_content_specification_id,
                        }
                    }
                },
            },
        }
        return computation_node

    @classmethod
    def _from_high_level(cls, id: str, name: str, node: SqliteComputationNode) -> Self:
        """
        Instantiate a `SqliteComputeNodeDefinition` from its high level representation.

        **Parameters**:
        - `id`: ID of the `SqliteComputeNode`.
        - `name`: Name of the `SqliteComputeNode`.
        - `node`: Pydantic model of the `SqliteComputeNode`.
        """
        sqlite_node = json.loads(node.model_dump_json())
        return cls(
            id=id,
            name=name,
            query=sqlite_node["statement"],
            dependencies=sqlite_node["dependencies"],
            enable_logs_on_error=sqlite_node["enableLogsOnError"],
            enable_logs_on_success=sqlite_node["enableLogsOnSuccess"],
        )

    def build(
        self,
        dcr_id: str,
        node_definition: NodeDefinition,
        client: Client,
        session: Session,
    ) -> SqliteComputeNode:
        """
        Construct a SqliteComputeNode from the Node Definition.

        **Parameters**:
        - `dcr_id`: ID of the DCR the node is a member of.
        - `node_definition`: Definition of the SQLite Compute Node.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        - `session`: The session with which to communicate with the enclave.
        """
        return SqliteComputeNode(
            name=self.name,
            dcr_id=dcr_id,
            query=self.query,
            client=client,
            session=session,
            node_definition=node_definition,
            dependencies=self.dependencies,
            enable_logs_on_error=self.enable_logs_on_error,
            enable_logs_on_success=self.enable_logs_on_success,
            id=self.id,
        )


class SqliteComputeNode(StructuredOutputNode):
    """
    Class representing an SQLite Computation node.
    """

    def __init__(
        self,
        id: str,
        name: str,
        dcr_id: str,
        query: str,
        client: Client,
        session: Session,
        node_definition: NodeDefinition,
        dependencies: Optional[List[str]] = None,
        enable_logs_on_error: bool = False,
        enable_logs_on_success: bool = False,
    ) -> None:
        """
        Initialise a `SqliteComputeNode`:

        **Parameters**:
        - `id`: ID of the `SqliteComputeNode`.
        - `name`: Name of the `SqliteComputeNode`.
        - `dcr_id`: ID of the DCR the node is a member of.
        - `query`: SQLite query string.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        - `session`: The session with which to communicate with the enclave.
        - `dependencies`: Nodes that the `SqliteComputeNode` depends on.
        - `enable_logs_on_error`: Enable logs in the event of an error.
        - `enable_logs_on_success`: Enable logs when the computation is successful.
        """
        super().__init__(
            name=name,
            client=client,
            session=session,
            dcr_id=dcr_id,
            id=id,
            definition=node_definition,
        )
        self.query = query
        self.dependencies = dependencies
        self.enable_logs_on_error = enable_logs_on_error
        self.enable_logs_on_success = enable_logs_on_success
        # SQLite computations use the Python worker under the hood.
        self.specification_id = "decentriq.python-ml-worker-32-64"
        self.static_content_specification_id = "decentriq.driver"

    def _get_computation_id(self) -> str:
        """
        Retrieve the ID of the node corresponding to the computation.
        """
        return f"{self.id}_container"



from .builder import AbMediaDcrBuilder
from .ab_media import AbMediaDcr, AbMediaDcrDefinition
from .version import AUDIENCE_BUILDER_SUPPORTED_VERSION
from .features import AbMediaDcrFeatures
from .computations import Computation
from .rule_based_builder import (
    RuleBasedAudienceBuilder,
    AudienceCombinatorBuilder
)
from .lookalike_audience_builder import (
    LookalikeAudienceBuilder,
    LookalikeAudienceDefinition,
)
from .audience_definitions import (
    FilterOperator,
    Filter,
    MatchOperator,
    AudienceFilters,
    CombineOperator,
    AudienceCombinator,
    RuleBasedAudienceDefinition,
    LookalikeAudienceDefinition,
    AdvertiserAudienceDefinition,
    AudienceStatus,
)
from .audience_statistics_definition import (
    LalAudienceStatistics,
)
from ..types import *

__pdoc__ = {
    "builder": False,
    "ab_media": False,
    "version": False,
    "features": False,
    "computations": False,
}

__all__ = ["AbMediaDcrBuilder", "AbMediaDcr", "AbMediaDcrDefinition"]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: compute_container.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'compute_container.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17\x63ompute_container.proto\x12\x11\x63ompute_container\"a\n\x1c\x43ontainerWorkerConfiguration\x12\x30\n\x06static\x18\x01 \x01(\x0b\x32\x1e.compute_container.StaticImageH\x00\x42\x0f\n\rconfiguration\"\xbf\x03\n\x0bStaticImage\x12\x0f\n\x07\x63ommand\x18\x01 \x03(\t\x12\x32\n\x0bmountPoints\x18\x02 \x03(\x0b\x32\x1d.compute_container.MountPoint\x12\x12\n\noutputPath\x18\x03 \x01(\t\x12#\n\x1bincludeContainerLogsOnError\x18\x04 \x01(\x08\x12%\n\x1dincludeContainerLogsOnSuccess\x18\x05 \x01(\x08\x12\'\n\x1aminimumContainerMemorySize\x18\x06 \x01(\x04H\x00\x88\x01\x01\x12\x36\n)extraChunkCacheSizeToAvailableMemoryRatio\x18\x07 \x01(\x02H\x01\x88\x01\x01\x12\x46\n\x12proxyConfiguration\x18\x08 \x01(\x0b\x32%.compute_container.ProxyConfigurationH\x02\x88\x01\x01\x42\x1d\n\x1b_minimumContainerMemorySizeB,\n*_extraChunkCacheSizeToAvailableMemoryRatioB\x15\n\x13_proxyConfiguration\"\xb7\x01\n\x12ProxyConfiguration\x12?\n\x07liberal\x18\x01 \x01(\x0b\x32,.compute_container.ProxyConfigurationLiberalH\x00\x12O\n\x0f\x64omainAllowlist\x18\x02 \x01(\x0b\x32\x34.compute_container.ProxyConfigurationDomainAllowlistH\x00\x42\x0f\n\rconfiguration\"\x1b\n\x19ProxyConfigurationLiberal\"3\n!ProxyConfigurationDomainAllowlist\x12\x0e\n\x06\x64omain\x18\x01 \x03(\t\".\n\nMountPoint\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x12\n\ndependency\x18\x02 \x01(\tb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'compute_container_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_CONTAINERWORKERCONFIGURATION']._serialized_start=46
  _globals['_CONTAINERWORKERCONFIGURATION']._serialized_end=143
  _globals['_STATICIMAGE']._serialized_start=146
  _globals['_STATICIMAGE']._serialized_end=593
  _globals['_PROXYCONFIGURATION']._serialized_start=596
  _globals['_PROXYCONFIGURATION']._serialized_end=779
  _globals['_PROXYCONFIGURATIONLIBERAL']._serialized_start=781
  _globals['_PROXYCONFIGURATIONLIBERAL']._serialized_end=808
  _globals['_PROXYCONFIGURATIONDOMAINALLOWLIST']._serialized_start=810
  _globals['_PROXYCONFIGURATIONDOMAINALLOWLIST']._serialized_end=861
  _globals['_MOUNTPOINT']._serialized_start=863
  _globals['_MOUNTPOINT']._serialized_end=909
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: secret_store.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'secret_store.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12secret_store.proto\x12\x10secret_store_api\"2\n\x10SecretStoreEntry\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\x12\r\n\x05state\x18\x02 \x01(\x0c\"I\n\x13\x43reateSecretRequest\x12\x32\n\x06secret\x18\x01 \x01(\x0b\x32\".secret_store_api.SecretStoreEntry\"\"\n\x14\x43reateSecretResponse\x12\n\n\x02id\x18\x01 \x01(\t\"/\n\x10GetSecretRequest\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\r\"Y\n\x11GetSecretResponse\x12\x32\n\x06secret\x18\x01 \x01(\x0b\x32\".secret_store_api.SecretStoreEntry\x12\x10\n\x08\x63\x61sIndex\x18\x02 \x01(\x04\";\n\x13RemoveSecretRequest\x12\n\n\x02id\x18\x01 \x01(\t\x12\x18\n\x10\x65xpectedCasIndex\x18\x02 \x01(\x04\"\'\n\x14RemoveSecretResponse\x12\x0f\n\x07removed\x18\x01 \x01(\x08\"_\n\x16UpdateSecretAclRequest\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0e\n\x06newAcl\x18\x02 \x01(\x0c\x12\x0f\n\x07version\x18\x03 \x01(\r\x12\x18\n\x10\x65xpectedCasIndex\x18\x04 \x01(\x04\"*\n\x17UpdateSecretAclResponse\x12\x0f\n\x07updated\x18\x01 \x01(\x08\"\x9b\x02\n\x12SecretStoreRequest\x12=\n\x0c\x63reateSecret\x18\x01 \x01(\x0b\x32%.secret_store_api.CreateSecretRequestH\x00\x12\x37\n\tgetSecret\x18\x02 \x01(\x0b\x32\".secret_store_api.GetSecretRequestH\x00\x12=\n\x0cremoveSecret\x18\x03 \x01(\x0b\x32%.secret_store_api.RemoveSecretRequestH\x00\x12\x43\n\x0fupdateSecretAcl\x18\x04 \x01(\x0b\x32(.secret_store_api.UpdateSecretAclRequestH\x00\x42\t\n\x07request\"\xa1\x02\n\x13SecretStoreResponse\x12>\n\x0c\x63reateSecret\x18\x01 \x01(\x0b\x32&.secret_store_api.CreateSecretResponseH\x00\x12\x38\n\tgetSecret\x18\x02 \x01(\x0b\x32#.secret_store_api.GetSecretResponseH\x00\x12>\n\x0cremoveSecret\x18\x03 \x01(\x0b\x32&.secret_store_api.RemoveSecretResponseH\x00\x12\x44\n\x0fupdateSecretAcl\x18\x04 \x01(\x0b\x32).secret_store_api.UpdateSecretAclResponseH\x00\x42\n\n\x08responseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'secret_store_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SECRETSTOREENTRY']._serialized_start=40
  _globals['_SECRETSTOREENTRY']._serialized_end=90
  _globals['_CREATESECRETREQUEST']._serialized_start=92
  _globals['_CREATESECRETREQUEST']._serialized_end=165
  _globals['_CREATESECRETRESPONSE']._serialized_start=167
  _globals['_CREATESECRETRESPONSE']._serialized_end=201
  _globals['_GETSECRETREQUEST']._serialized_start=203
  _globals['_GETSECRETREQUEST']._serialized_end=250
  _globals['_GETSECRETRESPONSE']._serialized_start=252
  _globals['_GETSECRETRESPONSE']._serialized_end=341
  _globals['_REMOVESECRETREQUEST']._serialized_start=343
  _globals['_REMOVESECRETREQUEST']._serialized_end=402
  _globals['_REMOVESECRETRESPONSE']._serialized_start=404
  _globals['_REMOVESECRETRESPONSE']._serialized_end=443
  _globals['_UPDATESECRETACLREQUEST']._serialized_start=445
  _globals['_UPDATESECRETACLREQUEST']._serialized_end=540
  _globals['_UPDATESECRETACLRESPONSE']._serialized_start=542
  _globals['_UPDATESECRETACLRESPONSE']._serialized_end=584
  _globals['_SECRETSTOREREQUEST']._serialized_start=587
  _globals['_SECRETSTOREREQUEST']._serialized_end=870
  _globals['_SECRETSTORERESPONSE']._serialized_start=873
  _globals['_SECRETSTORERESPONSE']._serialized_end=1162
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import json
from typing import Dict, Any, List, Optional, TYPE_CHECKING

from ..session import Session
from .computations import Computation
from decentriq_dcr_compiler.schemas import RequestedAudiencePayload

if TYPE_CHECKING:
    from ..client import Client


class GetAudiencesForPublisherComputation(Computation):
    def __init__(self, dcr_id: str, client: Client, session: Session) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)

    def node_id(self) -> str:
        return "get_audiences_for_publisher"

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(request_type="getAudiencesForPublisher")

    def get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> Dict[str, Any]:
        result = super().get_results_str_from_zip(
            "audiences.json", interval=interval, timeout=timeout
        )
        return json.loads(result)


class EstimateAudienceSizeForPublisherComputation(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
        generate_audience: RequestedAudiencePayload,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)
        self.generate_audience = generate_audience

    def node_id(self) -> str:
        return "estimate_audience_size_for_publisher"

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(
            request_type="estimateAudienceSizeForPublisher",
            generate_audience=self.generate_audience,
        )

    def get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> Dict[str, Any]:
        results = super().get_results_str_from_zip(
            "audience_size.json", interval=interval, timeout=timeout
        )
        return json.loads(results)


class EstimateAudienceSizeForPublisherLalComputation(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
        generate_audience: RequestedAudiencePayload,
        lal_audience: RequestedAudiencePayload,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)
        self.generate_audience = generate_audience
        self.lal_audience = lal_audience

    def node_id(self) -> str:
        return "estimate_audience_size_for_publisher_lal"

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(
            request_type="estimateAudienceSizeForPublisherLal",
            generate_audience=self.generate_audience,
            lal_audience=self.lal_audience,
        )

    def get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> Dict[str, Any]:
        results = super().get_results_str_from_zip(
            "audience_size.json", interval=interval, timeout=timeout
        )
        return json.loads(results)


class GetAudienceUserListForPublisherComputation(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
        generate_audience: RequestedAudiencePayload,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)
        self.generate_audience = generate_audience

    def node_id(self) -> str:
        return "get_audience_user_list_for_publisher"

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(
            request_type="getAudienceUserListForPublisher",
            generate_audience=self.generate_audience,
        )

    def get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> List[str]:
        result = super().get_results_str_from_zip(
            "audience_users.csv", interval=interval, timeout=timeout
        )
        return [line for line in result.split("\n") if line]


class GetAudienceUserListForPublisherLalComputation(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
        generate_audience: RequestedAudiencePayload,
        lal_audience: RequestedAudiencePayload,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)
        self.generate_audience = generate_audience
        self.lal_audience = lal_audience

    def node_id(self) -> str:
        return "get_audience_user_list_for_publisher_lal"

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(
            request_type="getAudienceUserListForPublisherLal",
            generate_audience=self.generate_audience,
            lal_audience=self.lal_audience,
        )

    def get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> List[str]:
        result = super().get_results_str_from_zip(
            "audience_users.csv", interval=interval, timeout=timeout
        )
        return [line for line in result.split("\n") if line]

class ComputeInsightsComputation(Computation):
    def __init__(self, dcr_id: str, client: Client, session: Session) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)

    def node_id(self) -> str:
        return "compute_insights"

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(request_type="computeInsights")

    def get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> Dict[str, Any]:
        result = super().get_results_str_from_zip(
            "segments.json", interval=interval, timeout=timeout
        )
        return json.loads(result)

def _get_validation_results(node, **kwargs) -> Optional[dict[str, Any]]:
    node.run()
    try:
        report_str = node.get_results_str_from_zip("validation-report.json", **kwargs)
        return json.loads(report_str)
    except Exception as e:
        if "Input file does not exist" in str(e):
            return None
        else:
            raise e

class GetMatchingValidationReport(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)

    def node_id(self) -> str:
        return "matching_validation_report"

    def run(self) -> None:
        super().run(
            request_type="getMatchingValidationReport",
        )

    def run_and_get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> Optional[dict[str, Any]]:
        return _get_validation_results(self, interval=interval, timeout=timeout)

class GetSegmentsValidationReport(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)

    def node_id(self) -> str:
        return "segments_validation_report"

    def run(self) -> None:
        super().run(
            request_type="getSegmentsValidationReport",
        )

    def run_and_get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> Optional[dict[str, Any]]:
        return _get_validation_results(self, interval=interval, timeout=timeout)

class GetDemographicsValidationReport(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)

    def node_id(self) -> str:
        return "demographics_validation_report"

    def run(self) -> None:
        super().run(
            request_type="getDemographicsValidationReport",
        )

    def run_and_get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> Optional[dict[str, Any]]:
        return _get_validation_results(self, interval=interval, timeout=timeout)



"""
.. include:: ../../../../decentriq_platform_docs/sql_getting_started.md
___
"""

__docformat__ = "restructuredtext"
__pdoc__ = {
    "compute": False,
    "helpers": False,
    "proto": False,
}

from .compute import SqlCompute, SqlSchemaVerifier
from .helpers import (
    TabularDataNodeBuilder,
    read_input_csv_file,
    read_input_csv_string,
    read_sql_query_result_as_string,
    upload_and_publish_tabular_dataset,
)
from .proto.compute_sql_pb2 import PrimitiveType

__all__ = [
    "SqlCompute",
    "TabularDataNodeBuilder",
    "PrimitiveType",
    "read_input_csv_file",
    "read_input_csv_string",
    "upload_and_publish_tabular_dataset",
    "read_sql_query_result_as_string",
    "SqlSchemaVerifier",
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: microsoft_dsp.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'microsoft_dsp.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13microsoft_dsp.proto\x12\rmicrosoft_dsp\"\x1a\n\nSingleFile\x12\x0c\n\x04name\x18\x01 \x01(\t\"\t\n\x07RawFile\"G\n\x07ZipFile\x12/\n\nsingleFile\x18\x01 \x01(\x0b\x32\x19.microsoft_dsp.SingleFileH\x00\x42\x0b\n\tselection\"\x83\x01\n\tSinkInput\x12\x12\n\ndependency\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12%\n\x03raw\x18\x03 \x01(\x0b\x32\x16.microsoft_dsp.RawFileH\x00\x12%\n\x03zip\x18\x04 \x01(\x0b\x32\x16.microsoft_dsp.ZipFileH\x00\x42\x06\n\x04\x66ile\"&\n\nMemberInfo\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\"/\n\x0bSegmentInfo\x12\x12\n\nshort_name\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\"\xac\x01\n\x1fMicrosoftDspWorkerConfiguration\x12\'\n\x05input\x18\x01 \x01(\x0b\x32\x18.microsoft_dsp.SinkInput\x12.\n\x0bmember_info\x18\x02 \x01(\x0b\x32\x19.microsoft_dsp.MemberInfo\x12\x30\n\x0csegment_info\x18\x03 \x01(\x0b\x32\x1a.microsoft_dsp.SegmentInfob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'microsoft_dsp_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SINGLEFILE']._serialized_start=38
  _globals['_SINGLEFILE']._serialized_end=64
  _globals['_RAWFILE']._serialized_start=66
  _globals['_RAWFILE']._serialized_end=75
  _globals['_ZIPFILE']._serialized_start=77
  _globals['_ZIPFILE']._serialized_end=148
  _globals['_SINKINPUT']._serialized_start=151
  _globals['_SINKINPUT']._serialized_end=282
  _globals['_MEMBERINFO']._serialized_start=284
  _globals['_MEMBERINFO']._serialized_end=322
  _globals['_SEGMENTINFO']._serialized_start=324
  _globals['_SEGMENTINFO']._serialized_end=371
  _globals['_MICROSOFTDSPWORKERCONFIGURATION']._serialized_start=374
  _globals['_MICROSOFTDSPWORKERCONFIGURATION']._serialized_end=546
# @@protoc_insertion_point(module_scope)



from .media import MediaDcr, CollaborationType
from .action import CreateLookalikeAudienceAction
from .audiences import Audiences
from .lookalike_audience_builder import LookalikeAudienceBuilder
from .remarketing_audience_builder import (
    RemarketingAudienceBuilder,
    RemarketingAudienceDefinition,
)
from .rule_based_builder import (
    RuleBasedAudienceBuilder,
    AudienceFilters,
    Filter,
    FilterOperator,
    CombineOperator,
    AudienceCombinator,
    RuleBasedAudienceDefinition,
    RuleBasedAudienceBuilder,
    MatchOperator,
    AudienceCombinatorBuilder,
)
from .audience_statistics_result import LalAudienceStatistics
from .participant import Participant
from .validation_reports import ValidationReports
from decentriq_dcr_compiler.schemas import Permission
from . import api
from ..types import MatchingId

__pdoc__ = {
    "action": False,
    "api": False,
    "audience_statistics_result": True,
    "audiences": True,
    "compute_job": True,
    "helper": False,
    "lookalike_audience_builder": True,
    "media": True,
    "participant": True,
    "permission": True,
    "remarketing_audience_builder": True,
    "rule_based_builder": True,
    "validation_reports": True,
}

__all__ = [
    "CollaborationType",
    "MediaDcr",
    "CreateLookalikeAudienceAction",
    "LookalikeAudienceBuilder",
    "RemarketingAudienceBuilder",
    "RemarketingAudienceDefinition",
    "Audiences",
    "RuleBasedAudienceBuilder",
    "AudienceFilters",
    "Filter",
    "FilterOperator",
    "CombineOperator",
    "AudienceCombinator",
    "RuleBasedAudienceDefinition",
    "MatchOperator",
    "AudienceCombinatorBuilder",
    "LalAudienceStatistics",
    "api",
    "Participant",
    "ValidationReports",
    "Permission",
    "MatchingId",
]



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict, List, Union

from ..session import Session
from .high_level_node import ComputationNode, DataNode

if TYPE_CHECKING:
    from ..client import Client


_valid_node_name_pattern = re.compile(
    r"^([a-zA-Z]|[a-zA-Z][a-zA-Z0-9_ -]*[a-zA-Z0-9])$"
)


class NodeDefinition(ABC):
    def __init__(self, name: str, id: str) -> None:
        # This check makes sure that node names do not cause issues when they
        # are used as file names within containers.
        if not _valid_node_name_pattern.match(name):
            raise Exception(
                f"Cannot create a node with name '{name}' as it contains invalid characters."
                " Node names should only contain alphanumeric characters,"
                " non-leading and non-trailing underscores/hyphens, as well as whitespace."
                " In addition, names should not start with a number."
            )
        super().__init__()
        self.id = id
        self.name = name

    @abstractmethod
    def _get_high_level_representation(self) -> Dict[str, str]:
        """
        Retrieve the high level representation of the `HighLevelNode`.
        """
        pass

    @abstractmethod
    def build(
        self,
        dcr_id: str,
        node_definition: NodeDefinition,
        client: Client,
        session: Session,
    ) -> Union[ComputationNode, DataNode]:
        pass

    @property
    @abstractmethod
    def required_workers(
        self,
    ) -> List[str]:
        pass

    def is_data_node(self) -> bool:
        hl = self._get_high_level_representation()
        assert "kind" in hl
        if "leaf" in hl["kind"]:
            return True
        return False

    def is_computation_node(self) -> bool:
        hl = self._get_high_level_representation()
        assert "kind" in hl
        if "computation" in hl["kind"]:
            return True
        return False


import hashlib
import hmac
from base64 import b64decode
from typing_extensions import Self
from typing import Any, Callable, List, Tuple, TypeVar

import chily
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF

from .api import Api, Endpoints
from .authentication import Auth, Sigma
from .graphql import GqlClient
from .proto.attestation_pb2 import AttestationSpecification, Fatquote
from .proto.data_room_pb2 import ComputeNodeProtocol
from .proto.delta_enclave_api_pb2 import DataNoncePubkey, Request, Response
from .proto.gcg_pb2 import GcgRequest, GcgRequestV2, GcgResponse, GcgResponseV2, Pki, UserAuth
from .proto.length_delimited import parse_length_delimited, serialize_length_delimited
from .verification import Verification
from .logger import logger


class EnclaveError(Exception):
    pass


CompilerRequest = TypeVar("CompilerRequest")
CompilerResponse = TypeVar("CompilerResponse")


def datanoncepubkey_to_message(
    encrypted_data: bytes,
    nonce: bytes,
    pubkey: bytes,
) -> DataNoncePubkey:
    message = DataNoncePubkey()
    message.data = encrypted_data
    message.nonce = nonce
    message.pubkey = pubkey
    return message


def message_to_datanoncepubkey(message: bytes) -> Tuple[bytes, bytes, bytes]:
    parsed_msg = DataNoncePubkey()
    parse_length_delimited(message, parsed_msg)
    return (parsed_msg.data, parsed_msg.nonce, parsed_msg.pubkey)


class Channel:
    """
    Class for managing the communication with an enclave.
    """

    session_id: str
    api: Api
    graphql_api: GqlClient
    enclave_public_key: Any
    driver_attestation_specification: AttestationSpecification
    driver_attestation_specification_hash: str
    channel_keypair: chily.Keypair

    def __init__(
        self,
        driver_attestation_specification: AttestationSpecification,
        api: Api,
        graphql_api: GqlClient,
        unsafe_disable_known_root_ca_check: bool = False,
    ):
        # Request a session with the enclave
        driver_attestation_specification_hash = hashlib.sha256(
            serialize_length_delimited(driver_attestation_specification)
        ).hexdigest()
        data = graphql_api.post(
            """
            mutation CreateSession($input: CreateSessionInput!) {
                session {
                    create(input: $input) {
                        record {
                            id
                        }
                    }
                }
            }
            """,
            {
                "input": {
                    "enclaveAttestationHash": driver_attestation_specification_hash,
                }
            },
        )
        session_id: str = data["session"]["create"]["record"]["id"]

        # Get fatquote from session and verify it
        data = graphql_api.post(
            """
            query GetFatquote($id: String!) {
                session(id: $id) {
                    fatquoteNewTcbLevel
                }
            }
            """,
            {
                "id": session_id,
            },
        )
        encoded_fatquote: str = data["session"]["fatquoteNewTcbLevel"]
        fatquote = Fatquote()
        fatquote_bytes_encoded = b64decode(encoded_fatquote)
        parse_length_delimited(fatquote_bytes_encoded, fatquote)
        verification = Verification(
            attestation_specification=driver_attestation_specification
        )
        if unsafe_disable_known_root_ca_check == True:
            logger.warning("WARNING: ROOT CHECK VERIFICATION DISABLED FOR CURRENT ENCLAVE SESSION")
            verification.disable_known_root_ca_check()

        # Extract enclave public key from report data
        report_data = verification.verify(fatquote)
        enclave_public_key_encoded = bytearray(report_data[:32])
        enclave_public_key = chily.PublicKey.from_bytes(enclave_public_key_encoded)

        self.api = api
        self.graphql_api = graphql_api
        self.session_id = session_id
        self.enclave_public_key = enclave_public_key
        self.driver_attestation_specification = driver_attestation_specification
        self.driver_attestation_specification_hash = (
            driver_attestation_specification_hash
        )
        self.channel_keypair = chily.Keypair.from_random()

    def _encrypt_and_encode_data(self, data: bytes) -> DataNoncePubkey:
        nonce = chily.Nonce.from_random()
        cipher = chily.Cipher(self.channel_keypair.secret, self.enclave_public_key)
        encrypted_data = cipher.encrypt("client sent session data", data, nonce)
        data_nonce_pubkey = datanoncepubkey_to_message(
            encrypted_data,
            nonce.bytes,
            self.channel_keypair.public_key.bytes,
        )
        return data_nonce_pubkey

    def _decode_and_decrypt_data(self, data: bytes) -> bytes:
        decoded_data, data_nonce, _ = message_to_datanoncepubkey(data)
        cipher = chily.Cipher(self.channel_keypair.secret, self.enclave_public_key)
        return cipher.decrypt(
            "client received session data",
            decoded_data,
            chily.Nonce.from_bytes(data_nonce),
        )

    def _get_message_auth(self, auth: Auth) -> UserAuth:
        shared_key = self.channel_keypair.secret.diffie_hellman(self.enclave_public_key).bytes
        hkdf = HKDF(
            algorithm=hashes.SHA512(), length=64, info=b"IdP KDF Context", salt=b""
        )
        mac_key = hkdf.derive(shared_key)
        mac_tag = hmac.digest(mac_key, auth.user_id.encode(), "sha512")
        public_keys = self.channel_keypair.public_key.bytes + self.enclave_public_key.bytes
        signature = auth.sign(public_keys)
        sigma_auth = Sigma(signature, mac_tag, auth)
        user_auth = UserAuth(
            pki=Pki(
                certChainPem=sigma_auth.get_cert_chain(),
                signature=sigma_auth.get_signature(),
                idMac=sigma_auth.get_mac_tag(),
            ),
            enclaveEndorsements=auth.get_enclave_endorsements(),
        )
        return user_auth

    def send_request_v2(
        self,
        request: GcgRequestV2,
    ) -> GcgResponseV2:
        serialized_request = serialize_length_delimited(
            self._encrypt_and_encode_data(
                serialize_length_delimited(request),
            )
        )
        url = Endpoints.SESSION_MESSAGES_V2.replace(":sessionId", self.session_id)
        with self.api.post(
            url,
            serialized_request,
            {"Content-type": "application/octet-stream", "Accept-Version": "2"},
        ) as response:
            enclave_response: bytes = response.content
        response_container = Response()
        parse_length_delimited(enclave_response, response_container)
        if response_container.HasField("unsuccessfulResponse"):
            raise Exception(response_container.unsuccessfulResponse)
        decrypted_response = self._decode_and_decrypt_data(
            response_container.successfulResponse,
        )
        response = GcgResponseV2()
        parse_length_delimited(decrypted_response, response)
        return response

    def send_request_raw(
        self,
        request: bytes,
        protocol: int,
    ) -> List[bytes]:
        gcg_protocol = serialize_length_delimited(ComputeNodeProtocol(version=protocol))
        serialized_request = serialize_length_delimited(
            Request(
                deltaRequest=self._encrypt_and_encode_data(
                    gcg_protocol + request,
                )
            )
        )
        url = Endpoints.SESSION_MESSAGES.replace(":sessionId", self.session_id)
        with self.api.post(
            url,
            serialized_request,
            {"Content-type": "application/octet-stream", "Accept-Version": "2"},
        ) as response:
            enclave_response: bytes = response.content

        responses: List[bytes] = []
        offset = 0
        while offset < len(enclave_response):
            response_container = Response()
            offset += parse_length_delimited(
                enclave_response[offset:], response_container
            )
            if response_container.HasField("unsuccessfulResponse"):
                raise Exception(response_container.unsuccessfulResponse)
            else:
                decrypted_response = self._decode_and_decrypt_data(
                    response_container.successfulResponse,
                )
                response_protocol = ComputeNodeProtocol()
                response_offset = parse_length_delimited(
                    decrypted_response, response_protocol
                )
                if response_protocol.version != protocol:
                    raise Exception(
                        "Different response protocol version than requested"
                    )
                responses.append(decrypted_response[response_offset:])
        return responses

    def send_compilable_request(
        self,
        compile_request: Callable[[CompilerRequest, Self], bytes],
        request: CompilerRequest,
        decompile_response: Callable[[List[bytes]], CompilerResponse],
        protocol: int,
    ) -> CompilerResponse:
        compiled_request = compile_request(request, self)
        responses = self.send_request_raw(compiled_request, protocol)
        response = decompile_response(responses)
        return response

    def send_request(
        self,
        request: GcgRequest,
        protocol: int,
        auth: Auth,
    ) -> List[GcgResponse]:
        """
        Low-level method for sending a raw `GcgRequest` to the enclave.
        Use this method if any of the convenience methods (such as `run_computation`) don't perform
        the exact task you want.
        """

        def parse_response(response_encoded: bytes) -> GcgResponse:
            response = GcgResponse()
            parse_length_delimited(response_encoded, response)
            if response.HasField("failure"):
                raise EnclaveError(response.failure)
            return response

        request.userAuth.CopyFrom(self._get_message_auth(auth))
        responses_encoded = self.send_request_raw(
            serialize_length_delimited(request), protocol
        )
        responses = list(map(parse_response, responses_encoded))
        return responses



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: release_policy.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'release_policy.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14release_policy.proto\x12\x12release_policy_api\"M\n\rReleasePolicy\x12\x31\n\x02v0\x18\x01 \x01(\x0b\x32#.release_policy_api.ReleasePolicyV0H\x00\x42\t\n\x07version\"\x93\x01\n\x0fReleasePolicyV0\x12\x36\n\x0csecretSource\x18\x01 \x01(\x0b\x32 .release_policy_api.SecretSource\x12H\n\x15\x61llowedComputeContext\x18\x02 \x01(\x0b\x32).release_policy_api.AllowedComputeContext\"U\n\x15\x41llowedComputeContext\x12\x34\n\x03\x64\x63r\x18\x01 \x01(\x0b\x32%.release_policy_api.DcrComputeContextH\x00\x42\x06\n\x04kind\";\n\x11\x44\x63rComputeContext\x12\x12\n\ndataRoomId\x18\x01 \x01(\t\x12\x12\n\ntargetLeaf\x18\x02 \x01(\t\"T\n\x0cSecretSource\x12<\n\x0bsecretStore\x18\x01 \x01(\x0b\x32%.release_policy_api.SecretStoreSourceH\x00\x42\x06\n\x04kind\"%\n\x11SecretStoreSource\x12\x10\n\x08secretId\x18\x01 \x01(\tb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'release_policy_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_RELEASEPOLICY']._serialized_start=44
  _globals['_RELEASEPOLICY']._serialized_end=121
  _globals['_RELEASEPOLICYV0']._serialized_start=124
  _globals['_RELEASEPOLICYV0']._serialized_end=271
  _globals['_ALLOWEDCOMPUTECONTEXT']._serialized_start=273
  _globals['_ALLOWEDCOMPUTECONTEXT']._serialized_end=358
  _globals['_DCRCOMPUTECONTEXT']._serialized_start=360
  _globals['_DCRCOMPUTECONTEXT']._serialized_end=419
  _globals['_SECRETSOURCE']._serialized_start=421
  _globals['_SECRETSOURCE']._serialized_end=505
  _globals['_SECRETSTORESOURCE']._serialized_start=507
  _globals['_SECRETSTORESOURCE']._serialized_end=544
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import google.protobuf.descriptor
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class ReleasePolicy(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    V0_FIELD_NUMBER: builtins.int
    @property
    def v0(self) -> global___ReleasePolicyV0: ...
    def __init__(
        self,
        *,
        v0: global___ReleasePolicyV0 | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["v0", b"v0", "version", b"version"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["v0", b"v0", "version", b"version"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["version", b"version"]) -> typing.Literal["v0"] | None: ...

global___ReleasePolicy = ReleasePolicy

@typing.final
class ReleasePolicyV0(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SECRETSOURCE_FIELD_NUMBER: builtins.int
    ALLOWEDCOMPUTECONTEXT_FIELD_NUMBER: builtins.int
    @property
    def secretSource(self) -> global___SecretSource:
        """Identifies the source of a secret.
        The policy creator must own this secret.
        """

    @property
    def allowedComputeContext(self) -> global___AllowedComputeContext:
        """/ Determines in which computation context a secret can be released."""

    def __init__(
        self,
        *,
        secretSource: global___SecretSource | None = ...,
        allowedComputeContext: global___AllowedComputeContext | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["allowedComputeContext", b"allowedComputeContext", "secretSource", b"secretSource"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["allowedComputeContext", b"allowedComputeContext", "secretSource", b"secretSource"]) -> None: ...

global___ReleasePolicyV0 = ReleasePolicyV0

@typing.final
class AllowedComputeContext(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DCR_FIELD_NUMBER: builtins.int
    @property
    def dcr(self) -> global___DcrComputeContext: ...
    def __init__(
        self,
        *,
        dcr: global___DcrComputeContext | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["dcr", b"dcr", "kind", b"kind"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["dcr", b"dcr", "kind", b"kind"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["kind", b"kind"]) -> typing.Literal["dcr"] | None: ...

global___AllowedComputeContext = AllowedComputeContext

@typing.final
class DcrComputeContext(google.protobuf.message.Message):
    """/ Identifies a particular leaf in a particular DCR"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATAROOMID_FIELD_NUMBER: builtins.int
    TARGETLEAF_FIELD_NUMBER: builtins.int
    dataRoomId: builtins.str
    targetLeaf: builtins.str
    def __init__(
        self,
        *,
        dataRoomId: builtins.str = ...,
        targetLeaf: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["dataRoomId", b"dataRoomId", "targetLeaf", b"targetLeaf"]) -> None: ...

global___DcrComputeContext = DcrComputeContext

@typing.final
class SecretSource(google.protobuf.message.Message):
    """/ Informs the enclave on how and where to get a particular secret
    / (such as an encryption key) from.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SECRETSTORE_FIELD_NUMBER: builtins.int
    @property
    def secretStore(self) -> global___SecretStoreSource: ...
    def __init__(
        self,
        *,
        secretStore: global___SecretStoreSource | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["kind", b"kind", "secretStore", b"secretStore"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["kind", b"kind", "secretStore", b"secretStore"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["kind", b"kind"]) -> typing.Literal["secretStore"] | None: ...

global___SecretSource = SecretSource

@typing.final
class SecretStoreSource(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SECRETID_FIELD_NUMBER: builtins.int
    secretId: builtins.str
    def __init__(
        self,
        *,
        secretId: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["secretId", b"secretId"]) -> None: ...

global___SecretStoreSource = SecretStoreSource



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Dict, List, Optional

from decentriq_dcr_compiler.schemas import SqlComputationNode
from typing_extensions import Self

from ..session import Session
from .high_level_node import StructuredOutputNode
from .node_definitions import NodeDefinition

if TYPE_CHECKING:
    from ..client import Client


class SqlComputeNodeDefinition(NodeDefinition):
    """
    Class representing an SQL Computation Node Definition.
    """

    def __init__(
        self,
        name: str,
        query: str,
        dependencies: Optional[List[str]] = None,
        minimum_rows_count: Optional[int] = None,
        id: Optional[str] = None,
    ) -> None:
        """
        Initialise a `SqlComputeNodeDefinition`:

        **Parameters**:
        `name`: Name of the `SqlComputeNodeDefinition`.
        `query`: SQL query string.
        `dependencies`: Node ids that the SQL node depends on.
        `minimum_rows_count`: Minimum number of rows required by the `SqlComputeNodeDefinition`.
        `id`: Optional ID of the `SqlComputeNodeDefinition`.
        """
        super().__init__(name=name, id=id or name)
        self.query = query
        self.dependencies = dependencies
        self.minimum_rows_count = minimum_rows_count
        self.specification_id = "decentriq.sql-worker"

    @property
    def required_workers(self):
        return [self.specification_id]

    def _get_high_level_representation(self) -> Dict[str, str]:
        """
        Retrieve the high level representation of the `SqlComputeNodeDefinition`.
        """
        dependencies = []
        if self.dependencies:
            for dependency in self.dependencies:
                dependencies.append(
                    {
                        "nodeId": dependency,
                        "tableName": dependency,
                    }
                )

        sql = {
            "dependencies": dependencies,
            "specificationId": self.specification_id,
            "statement": self.query,
        }
        if self.minimum_rows_count:
            sql["privacyFilter"]["minimumRowsCount"] = self.minimum_rows_count

        computation_node = {
            "id": self.id,
            "name": self.name,
            "kind": {
                "computation": {
                    "kind": {
                        "sql": sql,
                    }
                },
            },
        }
        return computation_node

    @classmethod
    def _from_high_level(cls, id: str, name: str, node: SqlComputationNode) -> Self:
        """
        Instantiate a `SqlComputeNodeDefinition` from its high level representation.

        **Parameters**:
        - `name`: Name of the `SqlComputeNodeDefinition`.
        - `node`: Pydantic model of the `SqlComputeNode`.
        """
        sql_node = json.loads(node.model_dump_json())
        minimum_rows_count = (
            None
            if not sql_node["privacyFilter"]
            else sql_node["privacyFilter"]["minimumRowsCount"]
        )
        return cls(
            id=id,
            name=name,
            query=sql_node["statement"],
            dependencies=sql_node["dependencies"],
            minimum_rows_count=minimum_rows_count,
        )

    def build(
        self,
        dcr_id: str,
        node_definition: NodeDefinition,
        client: Client,
        session: Session,
    ) -> SqlComputeNode:
        """
        Construct a SqlComputeNode from the Node Definition.

        **Parameters**:
        - `dcr_id`: ID of the DCR the node is a member of.
        - `node_definition`: Definition of the SQL Compute Node.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        - `session`: The session with which to communicate with the enclave.
        """
        return SqlComputeNode(
            name=self.name,
            dcr_id=dcr_id,
            query=self.query,
            client=client,
            session=session,
            node_definition=node_definition,
            dependencies=self.dependencies,
            minimum_rows_count=self.minimum_rows_count,
            id=self.id,
        )


class SqlComputeNode(StructuredOutputNode):
    """
    Class representing an SQL Computation Node.
    """

    def __init__(
        self,
        id: str,
        name: str,
        dcr_id: str,
        query: str,
        client: Client,
        session: Session,
        node_definition: NodeDefinition,
        *,
        dependencies: Optional[List[str]] = None,
        minimum_rows_count: Optional[int] = None,
    ) -> None:
        """
        Initialise a `SqlComputeNode`:

        **Parameters**:
        - `id`: ID of the `SqlComputeNode`.
        - `name`: Name of the `SqlComputeNode`.
        - `dcr_id`: ID of the DCR the node is a member of.
        - `query`: SQL query string.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        - `session`: The session with which to communicate with the enclave.
        - `dependencies`: Nodes that the `SqlComputeNode` depends on.
        - `minimum_rows_count`: Minimum number of rows required by the `SqlComputeNode`.
        """
        super().__init__(
            name=name,
            client=client,
            session=session,
            dcr_id=dcr_id,
            id=id,
            definition=node_definition,
        )
        self.statement = query
        self.dependencies = dependencies
        self.minimum_rows_count = minimum_rows_count
        self.specification_id = "decentriq.sql-worker"

    def _get_computation_id(self) -> str:
        """
        Retrieve the ID of the node corresponding to the computation.
        """
        return self.id



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class Fatquote(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EPID_FIELD_NUMBER: builtins.int
    DCAP_FIELD_NUMBER: builtins.int
    NITRO_FIELD_NUMBER: builtins.int
    SNP_FIELD_NUMBER: builtins.int
    DCAPMRSIGNER_FIELD_NUMBER: builtins.int
    @property
    def epid(self) -> global___FatquoteEpid: ...
    @property
    def dcap(self) -> global___FatquoteDcap: ...
    @property
    def nitro(self) -> global___FatquoteNitro: ...
    @property
    def snp(self) -> global___FatquoteSnp: ...
    @property
    def dcapMrsigner(self) -> global___FatquoteDcapMrsigner: ...
    def __init__(
        self,
        *,
        epid: global___FatquoteEpid | None = ...,
        dcap: global___FatquoteDcap | None = ...,
        nitro: global___FatquoteNitro | None = ...,
        snp: global___FatquoteSnp | None = ...,
        dcapMrsigner: global___FatquoteDcapMrsigner | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["dcap", b"dcap", "dcapMrsigner", b"dcapMrsigner", "epid", b"epid", "fatquote", b"fatquote", "nitro", b"nitro", "snp", b"snp"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["dcap", b"dcap", "dcapMrsigner", b"dcapMrsigner", "epid", b"epid", "fatquote", b"fatquote", "nitro", b"nitro", "snp", b"snp"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["fatquote", b"fatquote"]) -> typing.Literal["epid", "dcap", "nitro", "snp", "dcapMrsigner"] | None: ...

global___Fatquote = Fatquote

@typing.final
class FatquoteEpid(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IASRESPONSEBODY_FIELD_NUMBER: builtins.int
    IASCERTIFICATE_FIELD_NUMBER: builtins.int
    IASSIGNATURE_FIELD_NUMBER: builtins.int
    IASROOTCADER_FIELD_NUMBER: builtins.int
    iasResponseBody: builtins.bytes
    iasCertificate: builtins.bytes
    iasSignature: builtins.bytes
    iasRootCaDer: builtins.bytes
    def __init__(
        self,
        *,
        iasResponseBody: builtins.bytes = ...,
        iasCertificate: builtins.bytes = ...,
        iasSignature: builtins.bytes = ...,
        iasRootCaDer: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["iasCertificate", b"iasCertificate", "iasResponseBody", b"iasResponseBody", "iasRootCaDer", b"iasRootCaDer", "iasSignature", b"iasSignature"]) -> None: ...

global___FatquoteEpid = FatquoteEpid

@typing.final
class FatquoteDcap(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DCAPQUOTE_FIELD_NUMBER: builtins.int
    TCBINFO_FIELD_NUMBER: builtins.int
    QEIDENTITY_FIELD_NUMBER: builtins.int
    TCBSIGNCERT_FIELD_NUMBER: builtins.int
    QESIGNCERT_FIELD_NUMBER: builtins.int
    DCAPROOTCADER_FIELD_NUMBER: builtins.int
    dcapQuote: builtins.bytes
    tcbInfo: builtins.bytes
    qeIdentity: builtins.bytes
    tcbSignCert: builtins.bytes
    qeSignCert: builtins.bytes
    dcapRootCaDer: builtins.bytes
    def __init__(
        self,
        *,
        dcapQuote: builtins.bytes = ...,
        tcbInfo: builtins.bytes = ...,
        qeIdentity: builtins.bytes = ...,
        tcbSignCert: builtins.bytes = ...,
        qeSignCert: builtins.bytes = ...,
        dcapRootCaDer: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["dcapQuote", b"dcapQuote", "dcapRootCaDer", b"dcapRootCaDer", "qeIdentity", b"qeIdentity", "qeSignCert", b"qeSignCert", "tcbInfo", b"tcbInfo", "tcbSignCert", b"tcbSignCert"]) -> None: ...

global___FatquoteDcap = FatquoteDcap

@typing.final
class FatquoteDcapMrsigner(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DCAPQUOTE_FIELD_NUMBER: builtins.int
    TCBINFO_FIELD_NUMBER: builtins.int
    QEIDENTITY_FIELD_NUMBER: builtins.int
    TCBSIGNCERT_FIELD_NUMBER: builtins.int
    QESIGNCERT_FIELD_NUMBER: builtins.int
    DCAPROOTCADER_FIELD_NUMBER: builtins.int
    SIGSTOREEVIDENCE_FIELD_NUMBER: builtins.int
    dcapQuote: builtins.bytes
    tcbInfo: builtins.bytes
    qeIdentity: builtins.bytes
    tcbSignCert: builtins.bytes
    qeSignCert: builtins.bytes
    dcapRootCaDer: builtins.bytes
    @property
    def sigstoreEvidence(self) -> global___SigstoreEvidence: ...
    def __init__(
        self,
        *,
        dcapQuote: builtins.bytes = ...,
        tcbInfo: builtins.bytes = ...,
        qeIdentity: builtins.bytes = ...,
        tcbSignCert: builtins.bytes = ...,
        qeSignCert: builtins.bytes = ...,
        dcapRootCaDer: builtins.bytes = ...,
        sigstoreEvidence: global___SigstoreEvidence | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["sigstoreEvidence", b"sigstoreEvidence"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["dcapQuote", b"dcapQuote", "dcapRootCaDer", b"dcapRootCaDer", "qeIdentity", b"qeIdentity", "qeSignCert", b"qeSignCert", "sigstoreEvidence", b"sigstoreEvidence", "tcbInfo", b"tcbInfo", "tcbSignCert", b"tcbSignCert"]) -> None: ...

global___FatquoteDcapMrsigner = FatquoteDcapMrsigner

@typing.final
class SigstoreFile(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    CONTENTS_FIELD_NUMBER: builtins.int
    name: builtins.str
    contents: builtins.bytes
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        contents: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["contents", b"contents", "name", b"name"]) -> None: ...

global___SigstoreFile = SigstoreFile

@typing.final
class SigstoreEvidence(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    V1_FIELD_NUMBER: builtins.int
    @property
    def v1(self) -> global___SigstoreEvidenceV1: ...
    def __init__(
        self,
        *,
        v1: global___SigstoreEvidenceV1 | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["sigstore_evidence", b"sigstore_evidence", "v1", b"v1"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["sigstore_evidence", b"sigstore_evidence", "v1", b"v1"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["sigstore_evidence", b"sigstore_evidence"]) -> typing.Literal["v1"] | None: ...

global___SigstoreEvidence = SigstoreEvidence

@typing.final
class SigstoreEvidenceV1(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FILES_FIELD_NUMBER: builtins.int
    REKORLOGENTRYJSON_FIELD_NUMBER: builtins.int
    rekorLogEntryJson: builtins.bytes
    """Contains the inclusion proof and the leaf cosign certificate in the transparency log and ultimately the final SHA of the sgxs."""
    @property
    def files(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SigstoreFile]:
        """This list contains the root.json chain, targets.json and trusted_root.json
        (1.root.json, 2.root.json, .., X.root.json, Y.targets.json, trusted_root.json.)
        """

    def __init__(
        self,
        *,
        files: collections.abc.Iterable[global___SigstoreFile] | None = ...,
        rekorLogEntryJson: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["files", b"files", "rekorLogEntryJson", b"rekorLogEntryJson"]) -> None: ...

global___SigstoreEvidenceV1 = SigstoreEvidenceV1

@typing.final
class FatquoteNitro(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COSE_FIELD_NUMBER: builtins.int
    NITROROOTCADER_FIELD_NUMBER: builtins.int
    cose: builtins.bytes
    nitroRootCaDer: builtins.bytes
    def __init__(
        self,
        *,
        cose: builtins.bytes = ...,
        nitroRootCaDer: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["cose", b"cose", "nitroRootCaDer", b"nitroRootCaDer"]) -> None: ...

global___FatquoteNitro = FatquoteNitro

@typing.final
class FatquoteSnp(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REPORTBIN_FIELD_NUMBER: builtins.int
    AMDARKDER_FIELD_NUMBER: builtins.int
    AMDSEVDER_FIELD_NUMBER: builtins.int
    VCEKCRTDER_FIELD_NUMBER: builtins.int
    REPORTDATA_FIELD_NUMBER: builtins.int
    ROUGHTIMEPUBKEY_FIELD_NUMBER: builtins.int
    ROUGHTIMENONCE_FIELD_NUMBER: builtins.int
    SIGNEDTIMESTAMP_FIELD_NUMBER: builtins.int
    DECENTRIQDER_FIELD_NUMBER: builtins.int
    CHIPDER_FIELD_NUMBER: builtins.int
    reportBin: builtins.bytes
    amdArkDer: builtins.bytes
    amdSevDer: builtins.bytes
    vcekCrtDer: builtins.bytes
    reportData: builtins.bytes
    roughtimePubKey: builtins.bytes
    roughtimeNonce: builtins.bytes
    signedTimestamp: builtins.bytes
    decentriqDer: builtins.bytes
    chipDer: builtins.bytes
    def __init__(
        self,
        *,
        reportBin: builtins.bytes = ...,
        amdArkDer: builtins.bytes = ...,
        amdSevDer: builtins.bytes = ...,
        vcekCrtDer: builtins.bytes = ...,
        reportData: builtins.bytes = ...,
        roughtimePubKey: builtins.bytes = ...,
        roughtimeNonce: builtins.bytes = ...,
        signedTimestamp: builtins.bytes = ...,
        decentriqDer: builtins.bytes = ...,
        chipDer: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["amdArkDer", b"amdArkDer", "amdSevDer", b"amdSevDer", "chipDer", b"chipDer", "decentriqDer", b"decentriqDer", "reportBin", b"reportBin", "reportData", b"reportData", "roughtimeNonce", b"roughtimeNonce", "roughtimePubKey", b"roughtimePubKey", "signedTimestamp", b"signedTimestamp", "vcekCrtDer", b"vcekCrtDer"]) -> None: ...

global___FatquoteSnp = FatquoteSnp

@typing.final
class AttestationSpecification(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INTELEPID_FIELD_NUMBER: builtins.int
    INTELDCAP_FIELD_NUMBER: builtins.int
    AWSNITRO_FIELD_NUMBER: builtins.int
    AMDSNP_FIELD_NUMBER: builtins.int
    INTELDCAPMRSIGNER_FIELD_NUMBER: builtins.int
    @property
    def intelEpid(self) -> global___AttestationSpecificationIntelEpid: ...
    @property
    def intelDcap(self) -> global___AttestationSpecificationIntelDcap: ...
    @property
    def awsNitro(self) -> global___AttestationSpecificationAwsNitro: ...
    @property
    def amdSnp(self) -> global___AttestationSpecificationAmdSnp: ...
    @property
    def intelDcapMrsigner(self) -> global___AttestationSpecificationIntelDcapMrsigner: ...
    def __init__(
        self,
        *,
        intelEpid: global___AttestationSpecificationIntelEpid | None = ...,
        intelDcap: global___AttestationSpecificationIntelDcap | None = ...,
        awsNitro: global___AttestationSpecificationAwsNitro | None = ...,
        amdSnp: global___AttestationSpecificationAmdSnp | None = ...,
        intelDcapMrsigner: global___AttestationSpecificationIntelDcapMrsigner | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["amdSnp", b"amdSnp", "attestation_specification", b"attestation_specification", "awsNitro", b"awsNitro", "intelDcap", b"intelDcap", "intelDcapMrsigner", b"intelDcapMrsigner", "intelEpid", b"intelEpid"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["amdSnp", b"amdSnp", "attestation_specification", b"attestation_specification", "awsNitro", b"awsNitro", "intelDcap", b"intelDcap", "intelDcapMrsigner", b"intelDcapMrsigner", "intelEpid", b"intelEpid"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["attestation_specification", b"attestation_specification"]) -> typing.Literal["intelEpid", "intelDcap", "awsNitro", "amdSnp", "intelDcapMrsigner"] | None: ...

global___AttestationSpecification = AttestationSpecification

@typing.final
class AttestationSpecificationIntelEpid(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MRENCLAVE_FIELD_NUMBER: builtins.int
    IASROOTCADER_FIELD_NUMBER: builtins.int
    ACCEPTDEBUG_FIELD_NUMBER: builtins.int
    ACCEPTGROUPOUTOFDATE_FIELD_NUMBER: builtins.int
    ACCEPTCONFIGURATIONNEEDED_FIELD_NUMBER: builtins.int
    mrenclave: builtins.bytes
    iasRootCaDer: builtins.bytes
    acceptDebug: builtins.bool
    acceptGroupOutOfDate: builtins.bool
    acceptConfigurationNeeded: builtins.bool
    def __init__(
        self,
        *,
        mrenclave: builtins.bytes = ...,
        iasRootCaDer: builtins.bytes = ...,
        acceptDebug: builtins.bool = ...,
        acceptGroupOutOfDate: builtins.bool = ...,
        acceptConfigurationNeeded: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["acceptConfigurationNeeded", b"acceptConfigurationNeeded", "acceptDebug", b"acceptDebug", "acceptGroupOutOfDate", b"acceptGroupOutOfDate", "iasRootCaDer", b"iasRootCaDer", "mrenclave", b"mrenclave"]) -> None: ...

global___AttestationSpecificationIntelEpid = AttestationSpecificationIntelEpid

@typing.final
class AttestationSpecificationIntelDcap(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MRENCLAVE_FIELD_NUMBER: builtins.int
    DCAPROOTCADER_FIELD_NUMBER: builtins.int
    ACCEPTDEBUG_FIELD_NUMBER: builtins.int
    ACCEPTOUTOFDATE_FIELD_NUMBER: builtins.int
    ACCEPTCONFIGURATIONNEEDED_FIELD_NUMBER: builtins.int
    ACCEPTREVOKED_FIELD_NUMBER: builtins.int
    mrenclave: builtins.bytes
    dcapRootCaDer: builtins.bytes
    acceptDebug: builtins.bool
    acceptOutOfDate: builtins.bool
    acceptConfigurationNeeded: builtins.bool
    acceptRevoked: builtins.bool
    def __init__(
        self,
        *,
        mrenclave: builtins.bytes = ...,
        dcapRootCaDer: builtins.bytes = ...,
        acceptDebug: builtins.bool = ...,
        acceptOutOfDate: builtins.bool = ...,
        acceptConfigurationNeeded: builtins.bool = ...,
        acceptRevoked: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["acceptConfigurationNeeded", b"acceptConfigurationNeeded", "acceptDebug", b"acceptDebug", "acceptOutOfDate", b"acceptOutOfDate", "acceptRevoked", b"acceptRevoked", "dcapRootCaDer", b"dcapRootCaDer", "mrenclave", b"mrenclave"]) -> None: ...

global___AttestationSpecificationIntelDcap = AttestationSpecificationIntelDcap

@typing.final
class AttestationSpecificationAwsNitro(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NITROROOTCADER_FIELD_NUMBER: builtins.int
    PCR0_FIELD_NUMBER: builtins.int
    PCR1_FIELD_NUMBER: builtins.int
    PCR2_FIELD_NUMBER: builtins.int
    PCR8_FIELD_NUMBER: builtins.int
    nitroRootCaDer: builtins.bytes
    pcr0: builtins.bytes
    pcr1: builtins.bytes
    pcr2: builtins.bytes
    pcr8: builtins.bytes
    def __init__(
        self,
        *,
        nitroRootCaDer: builtins.bytes = ...,
        pcr0: builtins.bytes = ...,
        pcr1: builtins.bytes = ...,
        pcr2: builtins.bytes = ...,
        pcr8: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["nitroRootCaDer", b"nitroRootCaDer", "pcr0", b"pcr0", "pcr1", b"pcr1", "pcr2", b"pcr2", "pcr8", b"pcr8"]) -> None: ...

global___AttestationSpecificationAwsNitro = AttestationSpecificationAwsNitro

@typing.final
class AttestationSpecificationAmdSnp(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    AMDARKDER_FIELD_NUMBER: builtins.int
    MEASUREMENT_FIELD_NUMBER: builtins.int
    ROUGHTIMEPUBKEY_FIELD_NUMBER: builtins.int
    AUTHORIZEDCHIPIDS_FIELD_NUMBER: builtins.int
    DECENTRIQDER_FIELD_NUMBER: builtins.int
    amdArkDer: builtins.bytes
    measurement: builtins.bytes
    roughtimePubKey: builtins.bytes
    decentriqDer: builtins.bytes
    @property
    def authorizedChipIds(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    def __init__(
        self,
        *,
        amdArkDer: builtins.bytes = ...,
        measurement: builtins.bytes = ...,
        roughtimePubKey: builtins.bytes = ...,
        authorizedChipIds: collections.abc.Iterable[builtins.bytes] | None = ...,
        decentriqDer: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["amdArkDer", b"amdArkDer", "authorizedChipIds", b"authorizedChipIds", "decentriqDer", b"decentriqDer", "measurement", b"measurement", "roughtimePubKey", b"roughtimePubKey"]) -> None: ...

global___AttestationSpecificationAmdSnp = AttestationSpecificationAmdSnp

@typing.final
class AttestationSpecificationIntelDcapMrsigner(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MRSIGNER_FIELD_NUMBER: builtins.int
    DCAPROOTCADER_FIELD_NUMBER: builtins.int
    ACCEPTDEBUG_FIELD_NUMBER: builtins.int
    ACCEPTOUTOFDATE_FIELD_NUMBER: builtins.int
    ACCEPTCONFIGURATIONNEEDED_FIELD_NUMBER: builtins.int
    ACCEPTREVOKED_FIELD_NUMBER: builtins.int
    ISVPRODID_FIELD_NUMBER: builtins.int
    SIGSTOREROOTOLPCJSON_FIELD_NUMBER: builtins.int
    mrsigner: builtins.bytes
    dcapRootCaDer: builtins.bytes
    acceptDebug: builtins.bool
    acceptOutOfDate: builtins.bool
    acceptConfigurationNeeded: builtins.bool
    acceptRevoked: builtins.bool
    isvprodid: builtins.int
    sigstoreRootOlpcJson: builtins.bytes
    def __init__(
        self,
        *,
        mrsigner: builtins.bytes = ...,
        dcapRootCaDer: builtins.bytes = ...,
        acceptDebug: builtins.bool = ...,
        acceptOutOfDate: builtins.bool = ...,
        acceptConfigurationNeeded: builtins.bool = ...,
        acceptRevoked: builtins.bool = ...,
        isvprodid: builtins.int = ...,
        sigstoreRootOlpcJson: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["acceptConfigurationNeeded", b"acceptConfigurationNeeded", "acceptDebug", b"acceptDebug", "acceptOutOfDate", b"acceptOutOfDate", "acceptRevoked", b"acceptRevoked", "dcapRootCaDer", b"dcapRootCaDer", "isvprodid", b"isvprodid", "mrsigner", b"mrsigner", "sigstoreRootOlpcJson", b"sigstoreRootOlpcJson"]) -> None: ...

global___AttestationSpecificationIntelDcapMrsigner = AttestationSpecificationIntelDcapMrsigner



from typing import Optional, cast, TYPE_CHECKING
from ..client import Client
from ..types import MatchingId
from .data_lab import DataLab, DataLabConfig, ExistingDataLab


class DataLabBuilder:
    """
    A helper class to build a Data Lab.
    """

    def __init__(
        self,
        client: Client,
    ):
        self.name: str = ""
        self.has_demographics = False
        self.has_embeddings = False
        self.num_embeddings = 0
        self.has_segments = False
        self.matching_id = MatchingId.STRING
        self.validation_id: Optional[str] = None
        self.client = client
        self.existing = False
        self.data_lab_id: Optional[str] = None
        self._force_spark_validation = False
        self.drop_invalid_rows = True

    def _with_force_spark_validation(self, use_spark: bool):
        """
        Whether to force the data lab to use spark for checking
        whether datasets are valid. If this is false,
        the validation pipeline will decide based on the file size
        whether to use spark or not.

        **Parameters**:
        - `use_spark`: Force validation to use spark.
        """
        self._force_spark_validation = use_spark
        return self

    def with_name(self, name: str):
        """
        Set the name of the DataLab.

        **Parameters**:
        - `name`: Name to be used for the DataLab.
        """
        self.name = name

    def with_disable_drop_invalid_rows(self):
        """
        Disable dropping of invalid rows in the Data Lab.
        """
        self.drop_invalid_rows = False

    def with_matching_id_format(self, matching_id: MatchingId):
        """
        Set the matching ID format.

        **Parameters**:
        - `matching_id`: The type of matching ID to use.
        """
        self.matching_id = matching_id

    def with_demographics(self):
        """
        Enable demographics in the DataLab.
        """
        self.has_demographics = True

    def with_embeddings(self, num_embeddings: int):
        """
        Enable embeddings in the DataLab.

        **Parameters**:
        - `num_embeddings`: The number of embeddings the DataLab should use.
        """
        self.has_embeddings = True
        self.num_embeddings = num_embeddings

    def with_segments(self):
        self.has_segments = True

    def from_existing(self, data_lab_id: str):
        """
        Construct a new DataLab from an existing DataLab with the given ID.

        **Parameters**:
        - `data_lab_id`: The ID of the existing DataLab.
        """
        self.existing = True
        self.data_lab_id = data_lab_id

    def build(self) -> DataLab:
        """
        Build the DataLab.
        """
        if self.existing:
            # Build a new DataLab from an existing one.
            # The new DataLab will have the same configuration as the existing one.
            data_lab_definition = self.client.get_data_lab(cast(str, self.data_lab_id))
            cfg = DataLabConfig(
                data_lab_definition["name"],
                data_lab_definition["requireDemographicsDataset"],
                data_lab_definition["requireEmbeddingsDataset"],
                data_lab_definition["numEmbeddings"],
                data_lab_definition["requireSegmentsDataset"],
                data_lab_definition["matchingIdFormat"],
                force_spark_validation=data_lab_definition["forceSparkValidation"],
                drop_invalid_rows=data_lab_definition["dropInvalidRows"],
            )
            existing_data_lab = ExistingDataLab(data_lab_definition)
            return DataLab(self.client, cfg, existing_data_lab)
        else:
            # Build a new DataLab using the specified enclave specifications.
            cfg = DataLabConfig(
                self.name,
                self.has_demographics,
                self.has_embeddings,
                self.num_embeddings,
                self.has_segments,
                self.matching_id,
                force_spark_validation=self._force_spark_validation,
                drop_invalid_rows=self.drop_invalid_rows,
            )
            return DataLab(self.client, cfg)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: adform_dsp.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'adform_dsp.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x61\x64\x66orm_dsp.proto\x12\nadform_dsp\"\x1a\n\nSingleFile\x12\x0c\n\x04name\x18\x01 \x01(\t\"\t\n\x07RawFile\"D\n\x07ZipFile\x12,\n\nsingleFile\x18\x01 \x01(\x0b\x32\x16.adform_dsp.SingleFileH\x00\x42\x0b\n\tselection\"}\n\tSinkInput\x12\x12\n\ndependency\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\"\n\x03raw\x18\x03 \x01(\x0b\x32\x13.adform_dsp.RawFileH\x00\x12\"\n\x03zip\x18\x04 \x01(\x0b\x32\x13.adform_dsp.ZipFileH\x00\x42\x06\n\x04\x66ile\"q\n\x1c\x41\x64\x66ormDspWorkerConfiguration\x12$\n\x05input\x18\x01 \x01(\x0b\x32\x15.adform_dsp.SinkInput\x12\x16\n\x0esegment_owners\x18\x02 \x03(\t\x12\x13\n\x0bgzip_output\x18\x03 \x01(\x08\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'adform_dsp_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SINGLEFILE']._serialized_start=32
  _globals['_SINGLEFILE']._serialized_end=58
  _globals['_RAWFILE']._serialized_start=60
  _globals['_RAWFILE']._serialized_end=69
  _globals['_ZIPFILE']._serialized_start=71
  _globals['_ZIPFILE']._serialized_end=139
  _globals['_SINKINPUT']._serialized_start=141
  _globals['_SINKINPUT']._serialized_end=266
  _globals['_ADFORMDSPWORKERCONFIGURATION']._serialized_start=268
  _globals['_ADFORMDSPWORKERCONFIGURATION']._serialized_end=381
# @@protoc_insertion_point(module_scope)



from ..helper import flatten_mutable_fields


def test_flatten():
    input = {
        "version": "0.0.1",
        "advertiser_manifest_hash": None,
        "segments_manifest_hash": None,
        "embeddings_manifest_hash": None,
        "demographics_manifest_hash": None,
        "matching_manifest_hash": None,
        "audiences": [
            {
                "id": "overlap-shoes",
                "kind": "advertiser",
                "audience_type": "shoes",
                "audience_size": 2672,
                "mutable": {"name": "All publisher users in shoes", "status": "ready"},
            },
            {
                "id": "overlap-hot tubs",
                "kind": "advertiser",
                "audience_type": "hot tubs",
                "audience_size": 2716,
                "mutable": {
                    "name": "All publisher users in hot tubs",
                    "status": "ready",
                },
            },
            {
                "id": "overlap-insurance",
                "kind": "advertiser",
                "audience_type": "insurance",
                "audience_size": 2673,
                "mutable": {
                    "name": "All publisher users in insurance",
                    "status": "ready",
                },
            },
        ],
    }

    expected_output = {
        "version": "0.0.1",
        "advertiser_manifest_hash": None,
        "segments_manifest_hash": None,
        "embeddings_manifest_hash": None,
        "demographics_manifest_hash": None,
        "matching_manifest_hash": None,
        "audiences": [
            {
                "id": "overlap-shoes",
                "kind": "advertiser",
                "audience_type": "shoes",
                "audience_size": 2672,
                "name": "All publisher users in shoes",
                "status": "ready",
            },
            {
                "id": "overlap-hot tubs",
                "kind": "advertiser",
                "audience_type": "hot tubs",
                "audience_size": 2716,
                "name": "All publisher users in hot tubs",
                "status": "ready",
            },
            {
                "id": "overlap-insurance",
                "kind": "advertiser",
                "audience_type": "insurance",
                "audience_size": 2673,
                "name": "All publisher users in insurance",
                "status": "ready",
            },
        ],
    }

    result = flatten_mutable_fields(input)
    assert result == expected_output



from __future__ import annotations
import io
import json
from typing import Any, BinaryIO, Dict, TYPE_CHECKING, List, Optional, Union

from ..storage import Key
from .advertiser_computations import (
    GetAudiencesForAdvertiserComputation,
    EstimateAudienceSizeForAdvertiserComputation,
    GetAudienceUserListForAdvertiserComputation,
    GetAudienceUserListForAdvertiserLalComputation,
    RetrievePublishedDatasetsComputation,
    PublishAudiencesJson,
    EstimateAudienceSizeForAdvertiserLalComputation,
    GetLookalikeAudienceStatisticsComputation,
    GetDataAttributesComputation,
    GetAudiencesValidationReport,
    ComputeOverlapStatistics,
)
from .helper import (
    get_parameter_payloads,
    audience_depends_on_lookalike,
    get_dependencies,
)
from .audience_definitions import (
    RuleBasedAudienceDefinition,
    LookalikeAudienceDefinition,
    AdvertiserAudienceDefinition,
    AudienceDefinitions
)
from .audience_statistics_definition import LalAudienceStatistics
from decentriq_dcr_compiler.schemas import Audience2, Audience3
from decentriq_dcr_compiler.schemas import AbMediaRequest
from .request import Request
from .version import AUDIENCES_JSON_SUPPORTED_VERSION
from .audience_definitions import AudienceStatus

if TYPE_CHECKING:
    from .ab_media import AbMediaDcr


class Attribute:
    def __init__(self, name: str, values: List[str]) -> None:
        self.name = name
        self.values = values

    def as_dict(self) -> Dict[str, Any]:
        return {"name": self.name, "values": self.values}


class AdvertiserApi:
    """
    Provides the API for interacting with an Audience Builder DCR as an advertiser.
    """

    def __init__(self, dcr: AbMediaDcr) -> None:
        self.dcr_id = dcr.id
        self.session = dcr.session
        self.client = dcr.client
        self.features = dcr.features

    def upload_and_publish_audiences_dataset(
        self,
        data: BinaryIO,
        key: Key,
        file_name: str,
    ) -> None:
        """
        Upload the data to the Decentriq Platform and publish
        the data to the Audience Builder DCR.

        **Parameters**:
        - `data`: The data to upload and provision.
        - `key`: Key used to encrypt the file.
        - `file_name`: Name of the file.
        """
        audiences_dataset_id = self.client.upload_dataset(data, key, file_name)
        self.provision_audiences_dataset(audiences_dataset_id, key)

    def provision_audiences_dataset(self, dataset_id: str, key: Key) -> None:
        """
        Provision the audiences dataset to the Audience Builder DCR.

        **Parameters**:
        - `dataset_id`: ID of the uploaded dataset to provision to the DCR.
        - `key`: Key used to encrypt the dataset.
        """
        request_key = "publishAudiencesDataset"
        request = AbMediaRequest.model_validate(
            {
                request_key: {
                    "dataRoomIdHex": self.dcr_id,
                    "datasetHashHex": dataset_id,
                    "encryptionKeyHex": key.material.hex(),
                    "scopeIdHex": self.client._ensure_dcr_data_scope(self.dcr_id),
                },
            }
        )
        response = Request.send(request, self.session)
        if request_key not in response.model_dump_json():
            raise Exception(
                f"Failed to provision the audiences dataset to the Audience Builder DCR"
            )

    def deprovision_audiences_dataset(self) -> None:
        """
        Deprovision the audiences dataset from the Audience Builder DCR.
        """
        request_key = "unpublishAudiencesDataset"
        request = AbMediaRequest.model_validate(
            {
                request_key: {
                    "dataRoomIdHex": self.dcr_id,
                },
            }
        )
        response = Request.send(request, self.session)
        if request_key not in response.model_dump_json():
            raise Exception(
                f"Failed to deprovision the audiences dataset from the Audience Builder DCR"
            )

    def get_audiences(
        self,
    ) -> AudienceDefinitions:
        """
        Retrieve all advertiser audiences.
        """
        if (
            not self.features.has_enable_lookalike_audiences()
            and not self.features.has_enable_remarketing()
            and not self.features.has_enable_rule_based_audiences()
        ):
            raise Exception(
                "Unable to retrieve audiences. Lookalike, remarketing or rule-based features must be enabled."
            )

        audiences_dict = self._request_audiences()
        audiences = []
        for a in audiences_dict["audiences"]:
            if a["kind"] == "rulebased":
                audiences.append(
                    RuleBasedAudienceDefinition.from_dict(
                        a, audiences_dict["audiences"]
                    )
                )
            elif a["kind"] == "lookalike":
                audiences.append(
                    LookalikeAudienceDefinition.from_dict(
                        a, audiences_dict["audiences"]
                    )
                )
            elif a["kind"] == "advertiser":
                audiences.append(
                    AdvertiserAudienceDefinition.from_dict(
                        a, audiences_dict["audiences"]
                    )
                )
            else:
                raise Exception(f'Audience kind "{a["kind"]}" unknown')

        return audiences

    def _request_audiences(self) -> Dict[str, Any]:
        """
        Retrieve the audiences data structure as provided by the backend.
        """
        computation = GetAudiencesForAdvertiserComputation(
            dcr_id=self.dcr_id, client=self.client, session=self.session
        )
        return computation.run_and_get_results()

    def estimate_audience_size(self, audience_name: str) -> int:
        """
        Estimate the audience size for the advertiser.

        **Parameters**:
        - `audience_name`: Name of the audience for which the estimated audience size should be retrieved.
        """
        audiences_json = self._request_audiences()
        payloads = get_parameter_payloads(
            audience_name=audience_name, audiences_json=audiences_json
        )

        if audience_depends_on_lookalike(
            audience_name=audience_name, audiences_json=audiences_json
        ):
            computation = EstimateAudienceSizeForAdvertiserLalComputation(
                dcr_id=self.dcr_id,
                client=self.client,
                session=self.session,
                generate_audience=payloads.generate,
                lal_audience=payloads.lal,
            )
            result = computation.run_and_get_results()
            return result["audience_size"]
        else:
            computation = EstimateAudienceSizeForAdvertiserComputation(
                dcr_id=self.dcr_id,
                client=self.client,
                session=self.session,
                generate_audience=payloads.generate,
            )
            result = computation.run_and_get_results()
            return result["audience_size"]

    def get_validation_report(self) -> dict[str, Any]:
        """Get the validation reports for the advertiser datasets.

        Validation reports contain information about issues that were found
        in the provisioned data (e.g. values that don't match the expected schema
        or duplicated rows).

        **Returns**:
        A dictionary with a single key "audiences" under which the validation
        report is stored.
        """
        if not self.features.has_enable_drop_invalid_rows():
            raise Exception("This Audience Builder DCR does not validation of advertiser datasets.")
        return GetAudiencesValidationReport(
            dcr_id=self.dcr_id,
            client=self.client,
            session=self.session
        ).run_and_get_results()

    def get_audience_user_list(self, audience_name: str) -> List[str]:
        """
        Get the list of user IDs for the advertiser.

        **Parameters**:
        - `audience_name`: Name of the audience for which the audience user list should be retrieved.
        """
        if not self.features.has_enable_advertiser_audience_download():
            raise Exception(
                "This Audience Builder DCR does not support downloading audiences for advertisers"
            )

        audiences_json = self._request_audiences()
        payloads = get_parameter_payloads(
            audience_name=audience_name, audiences_json=audiences_json
        )

        if audience_depends_on_lookalike(
            audience_name=audience_name, audiences_json=audiences_json
        ):
            computation = GetAudienceUserListForAdvertiserLalComputation(
                dcr_id=self.dcr_id,
                client=self.client,
                session=self.session,
                generate_audience=payloads.generate,
                lal_audience=payloads.lal,
            )
            return computation.run_and_get_results()
        else:
            computation = GetAudienceUserListForAdvertiserComputation(
                dcr_id=self.dcr_id,
                client=self.client,
                session=self.session,
                generate_audience=payloads.generate,
            )
            return computation.run_and_get_results()

    def make_audiences_available_to_publisher(self, audience_names: List[str]):
        """
        Make the audiences with the given names available to the Publisher.
        The Publisher will be able to download the user ids belonging to these audiences.

        **Parameters**:
        - `audience_names`: Names of the audiences to make available to the publisher.
        """
        audiences_json = self._request_audiences()
        audiences = audiences_json.pop("audiences")
        audiences_dict = {a["mutable"]["name"]: a for a in audiences}

        for audience_name in audience_names:
            if audience_name not in audiences_dict:
                raise Exception(
                    f'Audience with name "{audience_name}" not in the list of available audiences'
                )

            audience_to_activate = audiences_dict[audience_name]

            has_remarketing = self.features.has_enable_remarketing()
            if not has_remarketing and audience_to_activate["kind"] == "advertiser":
                raise Exception(
                    f"The audience '{audience_name}' is of type 'advertiser', but this DCR does not allow remarketing."
                )
            else:
                # Set the status of the audience to "published".
                audience_to_activate["mutable"]["status"] = "published"
                audiences_dict[audience_name] = audience_to_activate

        # Update the audiences in the `audiences.json`.
        audiences_json["audiences"] = [a for a in audiences_dict.values()]
        self._upload_and_publish_audiences_json(audiences_json=audiences_json)

    def add_audiences(
        self,
        audiences: list[
            LookalikeAudienceDefinition
            | RuleBasedAudienceDefinition
        ],
    ):
        """
        Add the given audiences to the current DCR.
        The audiences will only be visible to the publisher after they have been published.

        **Parameters**:
        - `audiences`: The audience definitions to add.
        """

        # Get the existing audiences.
        audiences_json = self._request_audiences()
        existing_audiences = audiences_json.pop("audiences")
        audiences_dict = {}
        for a in existing_audiences:
            audience_name = a["mutable"]["name"]
            if audience_name in audiences_dict:
                raise Exception(f'Multiple audiences with name "{audience_name}"')
            audiences_dict[audience_name] = a
        existing_audiences_by_id = {a["id"]: a for a in audiences_dict.values()}

        # Add the new audiences.
        new_audiences_by_id = {a.id: a for a in audiences}
        all_audiences = existing_audiences + [new_audience.as_dict() for new_audience in audiences]
        for audience in audiences:
            if audience.kind == "advertiser":
                raise Exception("Advertiser audiences cannot be added through this function")

            if isinstance(audience, LookalikeAudienceDefinition):
                # Additional checks for lookalike audiences.
                if not self.features.has_enable_lookalike_audiences():
                    raise Exception(
                        "This Audience Builder DCR does not support Lookalike audiences"
                    )
                if audience.reach < 0 or audience.reach > 30:
                    raise Exception(
                        "Reach value must be an integer in the range 0 - 30"
                    )

            if audience.name in audiences_dict:
                raise Exception(
                    f'Audience with name "{audience.name}" already exists in audience list'
                )

            # Check if we need to mark any of the existing dependencies as PUBLISHED_AS_INTERMEDIATE
            # in order for them to be retrievable by the publisher.
            if audience.status == AudienceStatus.PUBLISHED:
                dependency_ids = get_dependencies(audience.id, all_audiences)
                for audience_id in dependency_ids:
                    if audience_id in existing_audiences_by_id:
                        dependency = existing_audiences_by_id[audience_id]
                        if dependency["mutable"]["status"] == AudienceStatus.READY.value:
                            dependency["mutable"]["status"] = AudienceStatus.PUBLISHED_AS_INTERMEDIATE.value
                    elif audience_id in new_audiences_by_id:
                        # Check whether the dependency is added as part of the same call.
                        # If yes, then it must have a published state.
                        dependency = new_audiences_by_id[audience_id]
                        if dependency.status == AudienceStatus.READY:
                            dependency.status = AudienceStatus.PUBLISHED_AS_INTERMEDIATE
                    else:
                        raise Exception(
                            f"Unable to find audience dependency '{audience_id}' in neither the existing audiences nor the audiences to be added"
                        )

        # Loop again over the audiences and dump them in JSON format.
        # The second loop is required since some of the audiences (that are dependencies
        # of other audiences to be added) got their status changed.
        for audience in audiences:
            parsed_audience = audience._as_ddc_audience()  # parse into the DDC representation
            audiences_dict[audience.name] = parsed_audience.model_dump(mode="json")

        # Update the audiences.json.
        audiences_json["audiences"] = [a for a in audiences_dict.values()]
        self._upload_and_publish_audiences_json(audiences_json=audiences_json)

    def get_audience_attributes(self) -> List[Attribute]:
        """
        Retrieve all audience attributes.
        These attributes can be used to build new rule-based audiences.
        """
        computation = GetDataAttributesComputation(
            dcr_id=self.dcr_id,
            client=self.client,
            session=self.session,
        )
        results = computation.run_and_get_results()
        attributes = [
            Attribute(k, v["values"]) for k, v in results["attributes"].items()
        ]
        return attributes

    def get_overlap_statistics(self) -> Dict[str, list[Any]]:
        """Get a report on how many users per audience were matched in the
        publisher's dataset. All numbers are rounded for privacy.

        **Returns**:
        A report with the following structure:

        ```
        {
            "overlap_statistics": [
                {
                    "audience_type": string,
                    "overlap_size": integer,
                    "andvertiser_size": integer
                }
            ]
        }
        ```

        If the DCR was configured to hide absolute audience counts, the
        values are omitted from the report.
        """
        return ComputeOverlapStatistics(
            dcr_id=self.dcr_id,
            client=self.client,
            session=self.session,
        ).run_and_get_results()

    def get_lookalike_audience_statistics(
        self, audience_name: str
    ) -> LalAudienceStatistics:
        """
        Retrieve lookalike statistics for an audience with the given name.

        **Parameters**:
        - `audience_name`: Name of the audience whose lookalike statistics should be retrieved.
        """
        audiences_json = self._request_audiences()
        if audience_depends_on_lookalike(
            audience_name=audience_name, audiences_json=audiences_json
        ):
            raise Exception("Requested audience cannot be a Lookalike audience")

        payloads = get_parameter_payloads(
            audience_name=audience_name, audiences_json=audiences_json
        )
        computation = GetLookalikeAudienceStatisticsComputation(
            dcr_id=self.dcr_id,
            client=self.client,
            session=self.session,
            generate_audience=payloads.generate,
        )
        results = computation.run_and_get_results()
        return LalAudienceStatistics.from_dict(results)

    def _update_manifest_hashes_in_audiences_json(self, audiences_json: Dict[str, Any]):
        # Set the dataset hashes to the latest versions.
        datasets = self._retrieve_published_datasets()
        if "matchingDatasetHashHex" not in datasets:
            raise Exception("Matching dataset must be provisioned to the DCR")
        audiences_json["matching_manifest_hash"] = datasets["matchingDatasetHashHex"]
        if "segmentsDatasetHashHex" in datasets:
            audiences_json["segments_manifest_hash"] = datasets[
                "segmentsDatasetHashHex"
            ]
        if "embeddingsDatasetHashHex" in datasets:
            audiences_json["embeddings_manifest_hash"] = datasets[
                "embeddingsDatasetHashHex"
            ]
        if "demographicsDatasetHashHex" in datasets:
            audiences_json["demographics_manifest_hash"] = datasets[
                "demographicsDatasetHashHex"
            ]
        if "audiencesDatasetHashHex" in datasets:
            audiences_json["advertiser_manifest_hash"] = datasets[
                "audiencesDatasetHashHex"
            ]

    def _upload_and_publish_audiences_json(self, audiences_json: Dict[str, Any]):
        audiences_json["version"] = AUDIENCES_JSON_SUPPORTED_VERSION

        # Update the manifest hashes whenever the `audiences.json` is published.
        self._update_manifest_hashes_in_audiences_json(audiences_json=audiences_json)

        key = Key()
        manifest_hash = self.client.upload_dataset(
            io.BytesIO(json.dumps(audiences_json).encode()),
            key,
            "audiences.json",
            is_accessory=True,
        )
        publish_audiences_json = PublishAudiencesJson(
            dcr_id=self.dcr_id,
            manifest_hash=manifest_hash,
            encryption_key_hex=key.material.hex(),
            client=self.client,
            session=self.session,
        )
        publish_audiences_json.run()

    def _retrieve_published_datasets(self) -> Dict[str, Any]:
        """
        Get all datasets that have been published to the DCR.
        """
        retrieve_published_datasets_computation = RetrievePublishedDatasetsComputation(
            dcr_id=self.dcr_id,
            client=self.client,
            session=self.session,
        )
        return retrieve_published_datasets_computation.run_and_get_results()



from google.protobuf.json_format import MessageToDict

from ...proto import (
    ComputeNodeFormat,
    parse_length_delimited,
    serialize_length_delimited,
)
from ..node import Node
from .proto import MetaSinkWorkerConfiguration, SinkInput

__docformat__ = "restructuredtext"
__pdoc__ = {
    "proto": False,
}


class MetaSink(Node):
    """
    Compute node that creates a custom audience on Meta and populates it with
    a list of user ids. The input file must contain a single column with one
    user id per line.
    """

    def __init__(
        self,
        name: str,
        input: SinkInput,
        access_token_dependency: str,
        ad_account_id: str,
        audience_name: str,
        api_version: str = "24.0",
    ) -> None:
        config = MetaSinkWorkerConfiguration(
            input=input,
            accessTokenDependency=access_token_dependency,
            adAccountId=ad_account_id,
            audienceName=audience_name,
            apiVersion=api_version,
        )
        config_serialized = serialize_length_delimited(config)
        dependencies = [input.dependency, access_token_dependency]
        super().__init__(
            name,
            config=config_serialized,
            enclave_type="decentriq.meta-sink-worker",
            dependencies=dependencies,
            output_format=ComputeNodeFormat.RAW,
        )


class MetaSinkWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = MetaSinkWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


__all__ = [
    "MetaSink",
]



from __future__ import annotations

import io
import json
import zipfile
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Dict, List, Optional, cast

from decentriq_dcr_compiler.schemas import TableLeafNodeV2
from typing_extensions import Self

from ..session import Session
from ..storage import Key
from .high_level_node import DataNode
from .node_definitions import NodeDefinition

if TYPE_CHECKING:
    from ..client import Client


class PrimitiveType(str, Enum):
    STRING = "string"
    INTEGER = "integer"
    FLOAT = "float"


class FormatType(str, Enum):
    STRING = "STRING"
    INTEGER = "INTEGER"
    FLOAT = "FLOAT"
    EMAIL = "EMAIL"
    DATE_ISO8601 = "DATE_ISO8601"
    PHONE_NUMBER_E164 = "PHONE_NUMBER_E164"
    HASH_SHA256_HEX = "HASH_SHA256_HEX"

    @staticmethod
    def from_primitive_type(tpe: str) -> FormatType:
        tpe = tpe.lower()
        if tpe == "integer":
            return FormatType.INTEGER
        elif tpe == "float":
            return FormatType.FLOAT
        elif tpe == "string":
            return FormatType.STRING
        else:
            raise Exception(f"Unable to convert data type {tpe} to a format type")

    @staticmethod
    def to_primitive_type(fmt: FormatType) -> str:
        if fmt == FormatType.INTEGER:
            return "integer"
        elif fmt == FormatType.FLOAT:
            return "float"
        elif fmt in {
            FormatType.STRING,
            FormatType.EMAIL,
            FormatType.DATE_ISO8601,
            FormatType.PHONE_NUMBER_E164,
            FormatType.HASH_SHA256_HEX,
        }:
            return "string"
        else:
            raise Exception(f"Unable to convert format type {fmt} to a primitive type")


class HashingAlgorithm(str, Enum):
    SHA256_HEX = "SHA256_HEX"


class NumericRangeRule:
    greater_than: Optional[float] = None
    greater_than_equals: Optional[float] = None
    less_than: Optional[float] = None
    less_than_equals: Optional[float] = None


@dataclass
class Column:
    format_type: FormatType
    name: str
    is_nullable: bool
    hash_with: Optional[HashingAlgorithm] = None
    in_range: Optional[NumericRangeRule] = None


class TableDataNodeDefinition(NodeDefinition):
    """
    Class representing a Table Data Node Definition.
    """

    def __init__(
        self,
        name: str,
        columns: List[Column],
        is_required: bool,
        id: Optional[str] = None,
        drop_invalid_rows: bool = False,
        unique_column_combinations: list[list[int]] = [],
    ) -> None:
        """
        Initialise a `TableDataNodeDefinition` instance.

        **Parameters**:
        - `name`: Name of the `TableDataNodeDefinition`
        - `columns`: Definition of the columns that make up the `TableDataNodeDefinition`.
        - `is_required`: Flag determining if the `RawDataNode` must be present for dependent computations.
        - `id`: Optional ID of the `TableDataNodeDefinition`
        - `drop_invalid_rows`: Whether to automatically drop rows that are marked as invalid during the validation step.
        - `unique_column_combinations`: Check that the given combination of
          columns are unique across the dataset. This should be a list of lists,
          where the inner lists list the 0-based column indices.
          For example, passing the list `[[0], [0, 1]]` would result in the enclave
          checking that all values in the first column are unique.
          It would further check that all tuples formed by the first and second
          column are unique.
        """
        super().__init__(name, id=id or name)
        self.is_required = is_required
        self.columns = columns
        self.specification_id = "decentriq.python-ml-worker-32-64"
        self.static_content_specification_id = "decentriq.driver"
        self.drop_invalid_rows = drop_invalid_rows
        self.unique_column_combinations = unique_column_combinations

    def _get_high_level_representation(self) -> Dict[str, str]:
        """
        Retrieve the high level representation of the `TableDataNodeDefinition`.
        """
        column_entries = []
        for column in self.columns:
            validation = {
                "name": column.name,
                "formatType": column.format_type.value,
                "allowNull": column.is_nullable,
            }
            if column.hash_with:
                validation["hashWith"] = column.hash_with.value
            if column.in_range:
                # TODO: fix this
                validation["inRange"] = column.in_range
            column_entries.append(
                {
                    "name": column.name,
                    "dataFormat": {
                        "isNullable": column.is_nullable,
                        "dataType": FormatType.to_primitive_type(column.format_type),
                    },
                    "validation": validation,
                }
            )

        table_validation = {}

        if self.unique_column_combinations:
            table_validation["uniqueness"] = {
                "uniqueKeys": [
                    {
                        "columns": indices
                    }
                ] for indices in self.unique_column_combinations
            }

        table_node = {
            "id": self.id,
            "name": self.name,
            "kind": {
                "leaf": {
                    "isRequired": self.is_required,
                    "kind": {
                        "table": {
                            "columns": column_entries,
                            "validationNode": {
                                "staticContentSpecificationId": self.static_content_specification_id,
                                "pythonSpecificationId": self.specification_id,
                                "validation": table_validation,
                                "dropInvalidRows": self.drop_invalid_rows,
                            },
                        }
                    },
                }
            },
        }
        return table_node

    @property
    def required_workers(self):
        return [self.static_content_specification_id, self.specification_id]

    @classmethod
    def _from_high_level(
        cls,
        id: str,
        name: str,
        node: TableLeafNodeV2,
        is_required: bool,
    ) -> Self:
        """
        Instantiate a `TableDataNodeDefinition` from its high level representation.

        **Parameters**:
        - `name`: Name of the `TableDataNodeDefinition`.
        - `node`: Pydantic model of the `TableDataNodeDefinition`.
        """
        node_dict = json.loads(node.model_dump_json())
        drop_invalid_rows = node_dict["validationNode"].get("dropInvalidRows", False)
        columns = [
            Column(
                name=column["name"],
                format_type=(
                    column["validation"]["formatType"]
                    if "formatType" in column.get("validation")
                    else FormatType.from_primitive_type(
                        column["dataFormat"]["dataType"]
                    )
                ),
                is_nullable=column["dataFormat"]["isNullable"],
                hash_with=column["validation"]["hashWith"],
                in_range=column["validation"]["inRange"],
            )
            for column in node_dict["columns"]
        ]
        return cls(
            id=id,
            name=name,
            columns=columns,
            is_required=is_required,
            drop_invalid_rows=drop_invalid_rows,
        )

    def build(
        self,
        dcr_id: str,
        node_definition: NodeDefinition,
        client: Client,
        session: Session,
    ) -> TableDataNode:
        """
        Construct a TableDataNode from the Node Definition.

        **Parameters**:
        - `dcr_id`: ID of the DCR the node is a member of.
        - `node_definition`: Definition of the Table Data Node.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        - `session`: The session with which to communicate with the enclave.
        """
        return TableDataNode(
            name=self.name,
            columns=self.columns,
            is_required=self.is_required,
            dcr_id=dcr_id,
            node_definition=cast(TableDataNodeDefinition, node_definition),
            client=client,
            session=session,
            id=self.id,
        )


class TableDataNode(DataNode):
    """
    Class representing a Table Data Node.
    """

    def __init__(
        self,
        id: str,
        name: str,
        columns: List[Column],
        is_required: bool,
        dcr_id: str,
        client: Client,
        session: Session,
        node_definition: TableDataNodeDefinition,
    ) -> None:
        """
        Initialise a `TableDataNode` instance.

        **Parameters**:
        - 'id': ID of the `TableDataNode`.
        - `name`: Name of the `TableDataNode`
        - `columns`: Definition of the columns that make up the `TableDataNode`.
        - `is_required`: Flag determining if the `TableDataNode` must be present for dependent computations.
        - `dcr_id`: ID of the DCR the node is a member of.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        - `session`: The session with which to communicate with the enclave.
        """
        super().__init__(
            name=name,
            is_required=is_required,
            dcr_id=dcr_id,
            client=client,
            session=session,
            id=id,
            definition=node_definition,
        )
        self.columns = columns

    # This data node needs to override the `super` implementation because
    # the leaf ID requires the "_leaf" suffix.
    def publish_dataset(self, manifest_hash: str, key: Key):
        """
        Publish data to the `TableDataNode`.

        **Parameters**:
        - `manifest_hash`: Hash identifying the dataset to be published.
        - `key`: Encryption key used to decrypt the dataset.
        """
        self.session.publish_dataset(
            self.dcr_id, manifest_hash, leaf_id=f"{self.id}_leaf", key=key
        )

    def get_validation_report_as_dict(self) -> Optional[Dict[str, str]]:
        """
        Retrieve the validation report corresponding to this `TableDataNode`.
        """
        validation_node_id = f"{self.id}_validation_report"
        result = self.session.run_computation_and_get_results(
            self.dcr_id, validation_node_id, interval=1
        )
        if result:
            validation_report = {}
            zip = zipfile.ZipFile(io.BytesIO(result), "r")
            if "validation-report.json" in zip.namelist():
                validation_report = json.loads(
                    zip.read("validation-report.json").decode()
                )
            return validation_report
        else:
            return None

    def remove_published_dataset(self) -> None:
        """
        Removes any dataset that is published to this node.
        """
        self.session.remove_published_dataset(self.dcr_id, f"{self.id}_leaf")



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import google.protobuf.descriptor
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class SnowflakeSource(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    WAREHOUSENAME_FIELD_NUMBER: builtins.int
    DATABASENAME_FIELD_NUMBER: builtins.int
    SCHEMANAME_FIELD_NUMBER: builtins.int
    TABLENAME_FIELD_NUMBER: builtins.int
    STAGENAME_FIELD_NUMBER: builtins.int
    warehouseName: builtins.str
    databaseName: builtins.str
    schemaName: builtins.str
    tableName: builtins.str
    stageName: builtins.str
    def __init__(
        self,
        *,
        warehouseName: builtins.str = ...,
        databaseName: builtins.str = ...,
        schemaName: builtins.str = ...,
        tableName: builtins.str = ...,
        stageName: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["databaseName", b"databaseName", "schemaName", b"schemaName", "stageName", b"stageName", "tableName", b"tableName", "warehouseName", b"warehouseName"]) -> None: ...

global___SnowflakeSource = SnowflakeSource

@typing.final
class DataSourceSnowflakeWorkerConfiguration(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SOURCE_FIELD_NUMBER: builtins.int
    CREDENTIALSDEPENDENCY_FIELD_NUMBER: builtins.int
    credentialsDependency: builtins.str
    @property
    def source(self) -> global___SnowflakeSource: ...
    def __init__(
        self,
        *,
        source: global___SnowflakeSource | None = ...,
        credentialsDependency: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["source", b"source"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["credentialsDependency", b"credentialsDependency", "source", b"source"]) -> None: ...

global___DataSourceSnowflakeWorkerConfiguration = DataSourceSnowflakeWorkerConfiguration
